# TODO: Update skipped tests to expect dictionary results from the execution
#       module functions.

import os.path
import random
import string

import pkg_resources  # pylint: disable=3rd-party-module-not-gated
import pytest
from pkg_resources import (  # pylint: disable=3rd-party-module-not-gated
    DistributionNotFound,
)

import salt.config
import salt.loader
import salt.modules.boto_vpc as boto_vpc
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.modules.boto_vpc import _maybe_set_name_tag, _maybe_set_tags
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

# pylint: disable=no-name-in-module,unused-import
try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto3
    from boto.exception import BotoServerError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    import moto
    from moto import mock_ec2_deprecated

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2_deprecated(self):
        """
        if the mock_ec2_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_vpc unit tests to use the @mock_ec2_deprecated decorator
        without a "NameError: name 'mock_ec2_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


# pylint: enable=no-name-in-module,unused-import

# the boto_vpc module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto_version = "2.8.0"
required_moto_version = "1.0.0"

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
cidr_block = "10.0.0.0/24"
dhcp_options_parameters = {
    "domain_name": "example.com",
    "domain_name_servers": ["1.2.3.4"],
    "ntp_servers": ["5.6.7.8"],
    "netbios_name_servers": ["10.0.0.1"],
    "netbios_node_type": 2,
}
network_acl_entry_parameters = ("fake", 100, -1, "allow", cidr_block)
dhcp_options_parameters.update(conn_parameters)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto.__version__) < Version(required_boto_version):
        return False
    else:
        return True


def _get_boto_version():
    """
    Returns the boto version
    """
    if not HAS_BOTO:
        return False
    return Version(boto.__version__)


def _get_moto_version():
    """
    Returns the moto version
    """
    try:
        return Version(str(moto.__version__))
    except AttributeError:
        try:
            return Version(pkg_resources.get_distribution("moto").version)
        except DistributionNotFound:
            return False


def _has_required_moto():
    """
    Returns True/False boolean depending on if Moto is installed and correct
    version.
    """
    if not HAS_MOTO:
        return False
    else:
        if _get_moto_version() < Version(required_moto_version):
            return False
        return True


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version() if HAS_BOTO else "None"
    ),
)
@pytest.mark.skipif(
    _has_required_moto() is False,
    reason="The moto version must be >= to version {}. Installed: {}".format(
        required_moto_version, _get_moto_version() if HAS_MOTO else "None"
    ),
)
class BotoVpcTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn3 = None

    def setup_loader_modules(self):
        self.opts = opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto", "boto3", "args", "systemd", "path", "platform"]
        )
        return {boto_vpc: {"__utils__": utils}}

    # Set up MagicMock to replace the boto3 session
    def setUp(self):
        super().setUp()
        boto_vpc.__init__(self.opts)
        delattr(self, "opts")
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn3 = MagicMock()
        self.addCleanup(delattr, self, "conn3")
        session_instance.client.return_value = self.conn3


class BotoVpcTestCaseMixin:
    conn = None

    def _create_vpc(self, name=None, tags=None):
        """
        Helper function to create a test vpc
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        vpc = self.conn.create_vpc(cidr_block)

        _maybe_set_name_tag(name, vpc)
        _maybe_set_tags(tags, vpc)
        return vpc

    def _create_subnet(
        self,
        vpc_id,
        cidr_block="10.0.0.0/26",
        name=None,
        tags=None,
        availability_zone=None,
    ):
        """
        Helper function to create a test subnet
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        subnet = self.conn.create_subnet(
            vpc_id, cidr_block, availability_zone=availability_zone
        )
        _maybe_set_name_tag(name, subnet)
        _maybe_set_tags(tags, subnet)
        return subnet

    def _create_internet_gateway(self, vpc_id, name=None, tags=None):
        """
        Helper function to create a test internet gateway
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        igw = self.conn.create_internet_gateway()
        _maybe_set_name_tag(name, igw)
        _maybe_set_tags(tags, igw)
        return igw

    def _create_customer_gateway(self, vpc_id, name=None, tags=None):
        """
        Helper function to create a test customer gateway
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        gw = self.conn.create_customer_gateway(vpc_id)
        _maybe_set_name_tag(name, gw)
        _maybe_set_tags(tags, gw)
        return gw

    def _create_dhcp_options(
        self,
        domain_name="example.com",
        domain_name_servers=None,
        ntp_servers=None,
        netbios_name_servers=None,
        netbios_node_type=2,
    ):
        """
        Helper function to create test dchp options
        """
        if not netbios_name_servers:
            netbios_name_servers = ["10.0.0.1"]
        if not ntp_servers:
            ntp_servers = ["5.6.7.8"]
        if not domain_name_servers:
            domain_name_servers = ["1.2.3.4"]

        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        return self.conn.create_dhcp_options(
            domain_name=domain_name,
            domain_name_servers=domain_name_servers,
            ntp_servers=ntp_servers,
            netbios_name_servers=netbios_name_servers,
            netbios_node_type=netbios_node_type,
        )

    def _create_network_acl(self, vpc_id):
        """
        Helper function to create test network acl
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        return self.conn.create_network_acl(vpc_id)

    def _create_network_acl_entry(
        self,
        network_acl_id,
        rule_number,
        protocol,
        rule_action,
        cidr_block,
        egress=None,
        icmp_code=None,
        icmp_type=None,
        port_range_from=None,
        port_range_to=None,
    ):
        """
        Helper function to create test network acl entry
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        return self.conn.create_network_acl_entry(
            network_acl_id,
            rule_number,
            protocol,
            rule_action,
            cidr_block,
            egress=egress,
            icmp_code=icmp_code,
            icmp_type=icmp_type,
            port_range_from=port_range_from,
            port_range_to=port_range_to,
        )

    def _create_route_table(self, vpc_id, name=None, tags=None):
        """
        Helper function to create a test route table
        """
        if not self.conn:
            self.conn = boto.vpc.connect_to_region(region)

        rtbl = self.conn.create_route_table(vpc_id)
        _maybe_set_name_tag(name, rtbl)
        _maybe_set_tags(tags, rtbl)
        return rtbl


class BotoVpcTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    """
    TestCase for salt.modules.boto_vpc module
    """

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_id_and_a_vpc_exists_the_vpc_exists_method_returns_true(
        self,
    ):
        """
        Tests checking vpc existence via id when the vpc already exists
        """
        vpc = self._create_vpc()

        vpc_exists_result = boto_vpc.exists(vpc_id=vpc.id, **conn_parameters)

        self.assertTrue(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_id_and_a_vpc_does_not_exist_the_vpc_exists_method_returns_false(
        self,
    ):
        """
        Tests checking vpc existence via id when the vpc does not exist
        """
        self._create_vpc()  # Created to ensure that the filters are applied correctly

        vpc_exists_result = boto_vpc.exists(vpc_id="fake", **conn_parameters)

        self.assertFalse(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_name_and_a_vpc_exists_the_vpc_exists_method_returns_true(
        self,
    ):
        """
        Tests checking vpc existence via name when vpc exists
        """
        self._create_vpc(name="test")

        vpc_exists_result = boto_vpc.exists(name="test", **conn_parameters)

        self.assertTrue(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_name_and_a_vpc_does_not_exist_the_vpc_exists_method_returns_false(
        self,
    ):
        """
        Tests checking vpc existence via name when vpc does not exist
        """
        self._create_vpc()  # Created to ensure that the filters are applied correctly

        vpc_exists_result = boto_vpc.exists(name="test", **conn_parameters)

        self.assertFalse(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_tags_and_a_vpc_exists_the_vpc_exists_method_returns_true(
        self,
    ):
        """
        Tests checking vpc existence via tag when vpc exists
        """
        self._create_vpc(tags={"test": "testvalue"})

        vpc_exists_result = boto_vpc.exists(
            tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertTrue(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_tags_and_a_vpc_does_not_exist_the_vpc_exists_method_returns_false(
        self,
    ):
        """
        Tests checking vpc existence via tag when vpc does not exist
        """
        self._create_vpc()  # Created to ensure that the filters are applied correctly

        vpc_exists_result = boto_vpc.exists(
            tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertFalse(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_cidr_and_a_vpc_exists_the_vpc_exists_method_returns_true(
        self,
    ):
        """
        Tests checking vpc existence via cidr when vpc exists
        """
        self._create_vpc()

        vpc_exists_result = boto_vpc.exists(cidr="10.0.0.0/24", **conn_parameters)

        self.assertTrue(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_by_cidr_and_a_vpc_does_not_exist_the_vpc_exists_method_returns_false(
        self,
    ):
        """
        Tests checking vpc existence via cidr when vpc does not exist
        """
        self._create_vpc()  # Created to ensure that the filters are applied correctly

        vpc_exists_result = boto_vpc.exists(cidr="10.10.10.10/24", **conn_parameters)

        self.assertFalse(vpc_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_vpc_exists_but_providing_no_filters_the_vpc_exists_method_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests checking vpc existence when no filters are provided
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following "
            "must be provided: vpc_id, vpc_name, "
            "cidr or tags.",
        ):
            boto_vpc.exists(**conn_parameters)

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_name(self):
        """
        Tests getting vpc id when filtering by name
        """
        vpc = self._create_vpc(name="test")

        get_id_result = boto_vpc.get_id(name="test", **conn_parameters)

        self.assertEqual(vpc.id, get_id_result["id"])

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_invalid_name(self):
        """
        Tests getting vpc id when filtering by invalid name
        """
        self._create_vpc(name="test")

        get_id_result = boto_vpc.get_id(name="test_fake", **conn_parameters)

        self.assertEqual(get_id_result["id"], None)

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_cidr(self):
        """
        Tests getting vpc id when filtering by cidr
        """
        vpc = self._create_vpc()

        get_id_result = boto_vpc.get_id(cidr="10.0.0.0/24", **conn_parameters)

        self.assertEqual(vpc.id, get_id_result["id"])

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_invalid_cidr(self):
        """
        Tests getting vpc id when filtering by invalid cidr
        """
        self._create_vpc()

        get_id_result = boto_vpc.get_id(cidr="10.10.10.10/24", **conn_parameters)

        self.assertEqual(get_id_result["id"], None)

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_tags(self):
        """
        Tests getting vpc id when filtering by tags
        """
        vpc = self._create_vpc(tags={"test": "testvalue"})

        get_id_result = boto_vpc.get_id(tags={"test": "testvalue"}, **conn_parameters)

        self.assertEqual(vpc.id, get_id_result["id"])

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_filtering_by_invalid_tags(self):
        """
        Tests getting vpc id when filtering by invalid tags
        """
        self._create_vpc(tags={"test": "testvalue"})

        get_id_result = boto_vpc.get_id(
            tags={"test": "fake-testvalue"}, **conn_parameters
        )

        self.assertEqual(get_id_result["id"], None)

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_not_providing_filters_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests getting vpc id but providing no filters
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following must be provided: vpc_name, cidr or tags.",
        ):
            boto_vpc.get_id(**conn_parameters)

    @mock_ec2_deprecated
    def test_get_vpc_id_method_when_more_than_one_vpc_is_matched_raises_a_salt_command_execution_error(
        self,
    ):
        """
        Tests getting vpc id but providing no filters
        """
        vpc1 = self._create_vpc(name="vpc-test1")
        vpc2 = self._create_vpc(name="vpc-test2")

        with self.assertRaisesRegex(
            CommandExecutionError, "Found more than one VPC matching the criteria."
        ):
            boto_vpc.get_id(cidr="10.0.0.0/24", **conn_parameters)

    @mock_ec2_deprecated
    def test_that_when_creating_a_vpc_succeeds_the_create_vpc_method_returns_true(self):
        """
        tests True VPC created.
        """
        vpc_creation_result = boto_vpc.create(cidr_block, **conn_parameters)

        self.assertTrue(vpc_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_a_vpc_and_specifying_a_vpc_name_succeeds_the_create_vpc_method_returns_true(
        self,
    ):
        """
        tests True VPC created.
        """
        vpc_creation_result = boto_vpc.create(
            cidr_block, vpc_name="test", **conn_parameters
        )

        self.assertTrue(vpc_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_a_vpc_and_specifying_tags_succeeds_the_create_vpc_method_returns_true(
        self,
    ):
        """
        tests True VPC created.
        """
        vpc_creation_result = boto_vpc.create(
            cidr_block, tags={"test": "value"}, **conn_parameters
        )

        self.assertTrue(vpc_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_creating_a_vpc_fails_the_create_vpc_method_returns_false(self):
        """
        tests False VPC not created.
        """
        with patch(
            "moto.ec2.models.VPCBackend.create_vpc",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            vpc_creation_result = boto_vpc.create(cidr_block, **conn_parameters)
            self.assertFalse(vpc_creation_result["created"])
            self.assertTrue("error" in vpc_creation_result)

    @mock_ec2_deprecated
    def test_that_when_deleting_an_existing_vpc_the_delete_vpc_method_returns_true(
        self,
    ):
        """
        Tests deleting an existing vpc
        """
        vpc = self._create_vpc()

        vpc_deletion_result = boto_vpc.delete(vpc.id, **conn_parameters)

        self.assertTrue(vpc_deletion_result)

    @mock_ec2_deprecated
    def test_that_when_deleting_a_non_existent_vpc_the_delete_vpc_method_returns_false(
        self,
    ):
        """
        Tests deleting a non-existent vpc
        """
        delete_vpc_result = boto_vpc.delete("1234", **conn_parameters)

        self.assertFalse(delete_vpc_result["deleted"])

    @mock_ec2_deprecated
    def test_that_when_describing_vpc_by_id_it_returns_the_dict_of_properties_returns_true(
        self,
    ):
        """
        Tests describing parameters via vpc id if vpc exist
        """
        # With moto 0.4.25 through 0.4.30, is_default is set to True.
        # 0.4.24 and older and 0.4.31 and newer, is_default is False
        if Version("0.4.25") <= _get_moto_version() < Version("0.4.31"):
            is_default = True
        else:
            is_default = False

        vpc = self._create_vpc(name="test", tags={"test": "testvalue"})

        describe_vpc = boto_vpc.describe(vpc_id=vpc.id, **conn_parameters)

        vpc_properties = dict(
            id=vpc.id,
            cidr_block=str(cidr_block),
            is_default=is_default,
            state="available",
            tags={"Name": "test", "test": "testvalue"},
            dhcp_options_id="dopt-7a8b9c2d",
            region="us-east-1",
            instance_tenancy="default",
        )

        self.assertEqual(describe_vpc, {"vpc": vpc_properties})

    @mock_ec2_deprecated
    def test_that_when_describing_vpc_by_id_it_returns_the_dict_of_properties_returns_false(
        self,
    ):
        """
        Tests describing parameters via vpc id if vpc does not exist
        """
        vpc = self._create_vpc(name="test", tags={"test": "testvalue"})

        describe_vpc = boto_vpc.describe(vpc_id="vpc-fake", **conn_parameters)

        self.assertFalse(describe_vpc["vpc"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_describing_vpc_by_id_on_connection_error_it_returns_error(self):
        """
        Tests describing parameters failure
        """
        vpc = self._create_vpc(name="test", tags={"test": "testvalue"})

        with patch(
            "moto.ec2.models.VPCBackend.get_all_vpcs",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            describe_result = boto_vpc.describe(vpc_id=vpc.id, **conn_parameters)
            self.assertTrue("error" in describe_result)

    @mock_ec2_deprecated
    def test_that_when_describing_vpc_but_providing_no_vpc_id_the_describe_method_returns_the_default_vpc(
        self,
    ):
        """
        Tests describing vpc without vpc id
        """
        describe_vpc = boto_vpc.describe(vpc_id=None, **conn_parameters)

        self.assertTrue(describe_vpc["vpc"]["is_default"])


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
@pytest.mark.skipif(
    _has_required_moto() is False,
    reason="The moto version must be >= to version {}".format(required_moto_version),
)
class BotoVpcSubnetsTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_get_subnet_association_single_subnet(self):
        """
        tests that given multiple subnet ids in the same VPC that the VPC ID is
        returned. The test is valuable because it uses a string as an argument
        to subnets as opposed to a list.
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)
        subnet_association = boto_vpc.get_subnet_association(
            subnets=subnet.id, **conn_parameters
        )
        self.assertEqual(vpc.id, subnet_association["vpc_id"])

    @mock_ec2_deprecated
    def test_get_subnet_association_multiple_subnets_same_vpc(self):
        """
        tests that given multiple subnet ids in the same VPC that the VPC ID is
        returned.
        """
        vpc = self._create_vpc()
        subnet_a = self._create_subnet(vpc.id, "10.0.0.0/25")
        subnet_b = self._create_subnet(vpc.id, "10.0.0.128/25")
        subnet_association = boto_vpc.get_subnet_association(
            [subnet_a.id, subnet_b.id], **conn_parameters
        )
        self.assertEqual(vpc.id, subnet_association["vpc_id"])

    @mock_ec2_deprecated
    def test_get_subnet_association_multiple_subnets_different_vpc(self):
        """
        tests that given multiple subnet ids in different VPCs that False is
        returned.
        """
        vpc_a = self._create_vpc()
        vpc_b = self.conn.create_vpc(cidr_block)
        subnet_a = self._create_subnet(vpc_a.id, "10.0.0.0/24")
        subnet_b = self._create_subnet(vpc_b.id, "10.0.0.0/24")
        subnet_association = boto_vpc.get_subnet_association(
            [subnet_a.id, subnet_b.id], **conn_parameters
        )
        self.assertEqual(set(subnet_association["vpc_ids"]), {vpc_a.id, vpc_b.id})

    @mock_ec2_deprecated
    def test_that_when_creating_a_subnet_succeeds_the_create_subnet_method_returns_true(
        self,
    ):
        """
        Tests creating a subnet successfully
        """
        vpc = self._create_vpc()

        subnet_creation_result = boto_vpc.create_subnet(
            vpc.id, "10.0.0.0/24", **conn_parameters
        )

        self.assertTrue(subnet_creation_result["created"])
        self.assertTrue("id" in subnet_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_a_subnet_and_specifying_a_name_succeeds_the_create_subnet_method_returns_true(
        self,
    ):
        """
        Tests creating a subnet successfully when specifying a name
        """
        vpc = self._create_vpc()

        subnet_creation_result = boto_vpc.create_subnet(
            vpc.id, "10.0.0.0/24", subnet_name="test", **conn_parameters
        )

        self.assertTrue(subnet_creation_result["created"])

    @mock_ec2_deprecated
    def test_that_when_creating_a_subnet_and_specifying_tags_succeeds_the_create_subnet_method_returns_true(
        self,
    ):
        """
        Tests creating a subnet successfully when specifying a tag
        """
        vpc = self._create_vpc()

        subnet_creation_result = boto_vpc.create_subnet(
            vpc.id, "10.0.0.0/24", tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertTrue(subnet_creation_result["created"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_creating_a_subnet_fails_the_create_subnet_method_returns_error(
        self,
    ):
        """
        Tests creating a subnet failure
        """
        vpc = self._create_vpc()

        with patch(
            "moto.ec2.models.SubnetBackend.create_subnet",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            subnet_creation_result = boto_vpc.create_subnet(
                vpc.id, "10.0.0.0/24", **conn_parameters
            )
            self.assertTrue("error" in subnet_creation_result)

    @mock_ec2_deprecated
    def test_that_when_deleting_an_existing_subnet_the_delete_subnet_method_returns_true(
        self,
    ):
        """
        Tests deleting an existing subnet
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        subnet_deletion_result = boto_vpc.delete_subnet(
            subnet_id=subnet.id, **conn_parameters
        )

        self.assertTrue(subnet_deletion_result["deleted"])

    @mock_ec2_deprecated
    def test_that_when_deleting_a_non_existent_subnet_the_delete_vpc_method_returns_false(
        self,
    ):
        """
        Tests deleting a subnet that doesn't exist
        """
        delete_subnet_result = boto_vpc.delete_subnet(
            subnet_id="1234", **conn_parameters
        )
        self.assertTrue("error" in delete_subnet_result)

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_subnet_exists_by_id_the_subnet_exists_method_returns_true(
        self,
    ):
        """
        Tests checking if a subnet exists when it does exist
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        subnet_exists_result = boto_vpc.subnet_exists(
            subnet_id=subnet.id, **conn_parameters
        )

        self.assertTrue(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_a_subnet_does_not_exist_the_subnet_exists_method_returns_false(
        self,
    ):
        """
        Tests checking if a subnet exists which doesn't exist
        """
        subnet_exists_result = boto_vpc.subnet_exists("fake", **conn_parameters)

        self.assertFalse(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_subnet_exists_by_name_the_subnet_exists_method_returns_true(
        self,
    ):
        """
        Tests checking subnet existence by name
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id, name="test")

        subnet_exists_result = boto_vpc.subnet_exists(name="test", **conn_parameters)

        self.assertTrue(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_subnet_exists_by_name_the_subnet_does_not_exist_the_subnet_method_returns_false(
        self,
    ):
        """
        Tests checking subnet existence by name when it doesn't exist
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id)

        subnet_exists_result = boto_vpc.subnet_exists(name="test", **conn_parameters)

        self.assertFalse(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_subnet_exists_by_tags_the_subnet_exists_method_returns_true(
        self,
    ):
        """
        Tests checking subnet existence by tag
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id, tags={"test": "testvalue"})

        subnet_exists_result = boto_vpc.subnet_exists(
            tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertTrue(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_checking_if_a_subnet_exists_by_tags_the_subnet_does_not_exist_the_subnet_method_returns_false(
        self,
    ):
        """
        Tests checking subnet existence by tag when subnet doesn't exist
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id)

        subnet_exists_result = boto_vpc.subnet_exists(
            tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertFalse(subnet_exists_result["exists"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_checking_if_a_subnet_exists_but_providing_no_filters_the_subnet_exists_method_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests checking subnet existence without any filters
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following must be specified: "
            "subnet id, cidr, subnet_name, tags, or zones.",
        ):
            boto_vpc.subnet_exists(**conn_parameters)

    @pytest.mark.skip(reason="Skip these tests while investigating failures")
    @mock_ec2_deprecated
    def test_that_describe_subnet_by_id_for_existing_subnet_returns_correct_data(self):
        """
        Tests describing a subnet by id.
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        describe_subnet_results = boto_vpc.describe_subnet(
            region=region, key=secret_key, keyid=access_key, subnet_id=subnet.id
        )
        self.assertEqual(
            set(describe_subnet_results["subnet"].keys()),
            {"id", "cidr_block", "availability_zone", "tags"},
        )

    @mock_ec2_deprecated
    def test_that_describe_subnet_by_id_for_non_existent_subnet_returns_none(self):
        """
        Tests describing a non-existent subnet by id.
        """
        self._create_vpc()

        describe_subnet_results = boto_vpc.describe_subnet(
            region=region, key=secret_key, keyid=access_key, subnet_id="subnet-a1b2c3"
        )
        self.assertEqual(describe_subnet_results["subnet"], None)

    @mock_ec2_deprecated
    def test_that_describe_subnet_by_name_for_existing_subnet_returns_correct_data(
        self,
    ):
        """
        Tests describing a subnet by name.
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id, name="test")

        describe_subnet_results = boto_vpc.describe_subnet(
            region=region, key=secret_key, keyid=access_key, subnet_name="test"
        )
        self.assertEqual(
            set(describe_subnet_results["subnet"].keys()),
            {"id", "vpc_id", "cidr_block", "availability_zone", "tags"},
        )

    @mock_ec2_deprecated
    def test_that_describe_subnet_by_name_for_non_existent_subnet_returns_none(self):
        """
        Tests describing a non-existent subnet by id.
        """
        self._create_vpc()

        describe_subnet_results = boto_vpc.describe_subnet(
            region=region, key=secret_key, keyid=access_key, subnet_name="test"
        )
        self.assertEqual(describe_subnet_results["subnet"], None)

    @mock_ec2_deprecated
    def test_that_describe_subnets_by_id_for_existing_subnet_returns_correct_data(self):
        """
        Tests describing multiple subnets by id.
        """
        vpc = self._create_vpc()
        subnet1 = self._create_subnet(vpc.id)
        subnet2 = self._create_subnet(vpc.id, cidr_block="10.0.0.64/26")

        describe_subnet_results = boto_vpc.describe_subnets(
            region=region,
            key=secret_key,
            keyid=access_key,
            subnet_ids=[subnet1.id, subnet2.id],
        )
        self.assertEqual(len(describe_subnet_results["subnets"]), 2)
        self.assertEqual(
            set(describe_subnet_results["subnets"][0].keys()),
            {"id", "vpc_id", "cidr_block", "availability_zone", "tags"},
        )

    @mock_ec2_deprecated
    def test_that_describe_subnets_by_name_for_existing_subnets_returns_correct_data(
        self,
    ):
        """
        Tests describing multiple subnets by id.
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id, name="subnet1")
        self._create_subnet(vpc.id, name="subnet2", cidr_block="10.0.0.64/26")

        describe_subnet_results = boto_vpc.describe_subnets(
            region=region,
            key=secret_key,
            keyid=access_key,
            subnet_names=["subnet1", "subnet2"],
        )
        self.assertEqual(len(describe_subnet_results["subnets"]), 2)
        self.assertEqual(
            set(describe_subnet_results["subnets"][0].keys()),
            {"id", "vpc_id", "cidr_block", "availability_zone", "tags"},
        )

    @mock_ec2_deprecated
    def test_create_subnet_passes_availability_zone(self):
        """
        Tests that the availability_zone kwarg is passed on to _create_resource
        """
        vpc = self._create_vpc()
        self._create_subnet(vpc.id, name="subnet1", availability_zone="us-east-1a")
        describe_subnet_results = boto_vpc.describe_subnets(
            region=region, key=secret_key, keyid=access_key, subnet_names=["subnet1"]
        )
        self.assertEqual(
            describe_subnet_results["subnets"][0]["availability_zone"], "us-east-1a"
        )


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
class BotoVpcInternetGatewayTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_that_when_creating_an_internet_gateway_the_create_internet_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an internet gateway successfully (with no vpc id or name)
        """

        igw_creation_result = boto_vpc.create_internet_gateway(
            region=region, key=secret_key, keyid=access_key
        )
        self.assertTrue(igw_creation_result.get("created"))

    @mock_ec2_deprecated
    def test_that_when_creating_an_internet_gateway_with_non_existent_vpc_the_create_internet_gateway_method_returns_an_error(
        self,
    ):
        """
        Tests that creating an internet gateway for a non-existent VPC fails.
        """

        igw_creation_result = boto_vpc.create_internet_gateway(
            region=region, key=secret_key, keyid=access_key, vpc_name="non-existent-vpc"
        )
        self.assertTrue("error" in igw_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_an_internet_gateway_with_vpc_name_specified_the_create_internet_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an internet gateway with vpc name specified.
        """

        self._create_vpc(name="test-vpc")

        igw_creation_result = boto_vpc.create_internet_gateway(
            region=region, key=secret_key, keyid=access_key, vpc_name="test-vpc"
        )

        self.assertTrue(igw_creation_result.get("created"))

    @mock_ec2_deprecated
    def test_that_when_creating_an_internet_gateway_with_vpc_id_specified_the_create_internet_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an internet gateway with vpc name specified.
        """

        vpc = self._create_vpc()

        igw_creation_result = boto_vpc.create_internet_gateway(
            region=region, key=secret_key, keyid=access_key, vpc_id=vpc.id
        )

        self.assertTrue(igw_creation_result.get("created"))


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
class BotoVpcNatGatewayTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_that_when_creating_an_nat_gateway_the_create_nat_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an nat gateway successfully (with subnet_id specified)
        """

        vpc = self._create_vpc()
        subnet = self._create_subnet(
            vpc.id, name="subnet1", availability_zone="us-east-1a"
        )
        ngw_creation_result = boto_vpc.create_nat_gateway(
            subnet_id=subnet.id, region=region, key=secret_key, keyid=access_key
        )
        self.assertTrue(ngw_creation_result.get("created"))

    @mock_ec2_deprecated
    def test_that_when_creating_an_nat_gateway_with_non_existent_subnet_the_create_nat_gateway_method_returns_an_error(
        self,
    ):
        """
        Tests that creating an nat gateway for a non-existent subnet fails.
        """

        ngw_creation_result = boto_vpc.create_nat_gateway(
            region=region,
            key=secret_key,
            keyid=access_key,
            subnet_name="non-existent-subnet",
        )
        self.assertTrue("error" in ngw_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_an_nat_gateway_with_subnet_name_specified_the_create_nat_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an nat gateway with subnet name specified.
        """

        vpc = self._create_vpc()
        subnet = self._create_subnet(
            vpc.id, name="test-subnet", availability_zone="us-east-1a"
        )
        ngw_creation_result = boto_vpc.create_nat_gateway(
            region=region, key=secret_key, keyid=access_key, subnet_name="test-subnet"
        )

        self.assertTrue(ngw_creation_result.get("created"))


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
class BotoVpcCustomerGatewayTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_customer_gateway_the_create_customer_gateway_method_returns_true(
        self,
    ):
        """
        Tests creating an internet gateway successfully (with no vpc id or name)
        """

        gw_creation_result = boto_vpc.create_customer_gateway(
            "ipsec.1", "10.1.1.1", None
        )
        self.assertTrue(gw_creation_result.get("created"))

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_checking_if_a_subnet_exists_by_id_the_subnet_exists_method_returns_true(
        self,
    ):
        """
        Tests checking if a subnet exists when it does exist
        """

        gw_creation_result = boto_vpc.create_customer_gateway(
            "ipsec.1", "10.1.1.1", None
        )
        gw_exists_result = boto_vpc.customer_gateway_exists(
            customer_gateway_id=gw_creation_result["id"]
        )
        self.assertTrue(gw_exists_result["exists"])

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_a_subnet_does_not_exist_the_subnet_exists_method_returns_false(
        self,
    ):
        """
        Tests checking if a subnet exists which doesn't exist
        """
        gw_exists_result = boto_vpc.customer_gateway_exists("fake")
        self.assertFalse(gw_exists_result["exists"])


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
@pytest.mark.skipif(
    _has_required_moto() is False,
    reason="The moto version must be >= to version {}".format(required_moto_version),
)
class BotoVpcDHCPOptionsTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_that_when_creating_dhcp_options_succeeds_the_create_dhcp_options_method_returns_true(
        self,
    ):
        """
        Tests creating dhcp options successfully
        """
        dhcp_options_creation_result = boto_vpc.create_dhcp_options(
            **dhcp_options_parameters
        )

        self.assertTrue(dhcp_options_creation_result["created"])

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_dhcp_options_and_specifying_a_name_succeeds_the_create_dhcp_options_method_returns_true(
        self,
    ):
        """
        Tests creating dchp options with name successfully
        """
        dhcp_options_creation_result = boto_vpc.create_dhcp_options(
            dhcp_options_name="test", **dhcp_options_parameters
        )

        self.assertTrue(dhcp_options_creation_result["created"])

    @mock_ec2_deprecated
    def test_that_when_creating_dhcp_options_and_specifying_tags_succeeds_the_create_dhcp_options_method_returns_true(
        self,
    ):
        """
        Tests creating dchp options with tag successfully
        """
        dhcp_options_creation_result = boto_vpc.create_dhcp_options(
            tags={"test": "testvalue"}, **dhcp_options_parameters
        )

        self.assertTrue(dhcp_options_creation_result["created"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_creating_dhcp_options_fails_the_create_dhcp_options_method_returns_error(
        self,
    ):
        """
        Tests creating dhcp options failure
        """
        with patch(
            "moto.ec2.models.DHCPOptionsSetBackend.create_dhcp_options",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            r = dhcp_options_creation_result = boto_vpc.create_dhcp_options(
                **dhcp_options_parameters
            )
            self.assertTrue("error" in r)

    @mock_ec2_deprecated
    def test_that_when_associating_an_existing_dhcp_options_set_to_an_existing_vpc_the_associate_dhcp_options_method_returns_true(
        self,
    ):
        """
        Tests associating existing dchp options successfully
        """
        vpc = self._create_vpc()
        dhcp_options = self._create_dhcp_options()

        dhcp_options_association_result = boto_vpc.associate_dhcp_options_to_vpc(
            dhcp_options.id, vpc.id, **conn_parameters
        )

        self.assertTrue(dhcp_options_association_result["associated"])

    @mock_ec2_deprecated
    def test_that_when_associating_a_non_existent_dhcp_options_set_to_an_existing_vpc_the_associate_dhcp_options_method_returns_error(
        self,
    ):
        """
        Tests associating non-existanct dhcp options successfully
        """
        vpc = self._create_vpc()

        dhcp_options_association_result = boto_vpc.associate_dhcp_options_to_vpc(
            "fake", vpc.id, **conn_parameters
        )

        self.assertTrue("error" in dhcp_options_association_result)

    @mock_ec2_deprecated
    def test_that_when_associating_an_existing_dhcp_options_set_to_a_non_existent_vpc_the_associate_dhcp_options_method_returns_false(
        self,
    ):
        """
        Tests associating existing dhcp options to non-existence vpc
        """
        dhcp_options = self._create_dhcp_options()

        dhcp_options_association_result = boto_vpc.associate_dhcp_options_to_vpc(
            dhcp_options.id, "fake", **conn_parameters
        )

        self.assertTrue("error" in dhcp_options_association_result)

    @mock_ec2_deprecated
    def test_that_when_creating_dhcp_options_set_to_an_existing_vpc_succeeds_the_associate_new_dhcp_options_method_returns_true(
        self,
    ):
        """
        Tests creation/association of dchp options to an existing vpc successfully
        """
        vpc = self._create_vpc()

        dhcp_creation_result = boto_vpc.create_dhcp_options(
            vpc_id=vpc.id, **dhcp_options_parameters
        )

        self.assertTrue(dhcp_creation_result["created"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_creating_and_associating_dhcp_options_set_to_an_existing_vpc_fails_creating_the_dhcp_options_the_associate_new_dhcp_options_method_raises_exception(
        self,
    ):
        """
        Tests creation failure during creation/association of dchp options to an existing vpc
        """
        vpc = self._create_vpc()

        with patch(
            "moto.ec2.models.DHCPOptionsSetBackend.create_dhcp_options",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            r = boto_vpc.associate_new_dhcp_options_to_vpc(
                vpc.id, **dhcp_options_parameters
            )
            self.assertTrue("error" in r)

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_creating_and_associating_dhcp_options_set_to_an_existing_vpc_fails_associating_the_dhcp_options_the_associate_new_dhcp_options_method_raises_exception(
        self,
    ):
        """
        Tests association failure during creation/association of dchp options to existing vpc
        """
        vpc = self._create_vpc()

        with patch(
            "moto.ec2.models.DHCPOptionsSetBackend.associate_dhcp_options",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            r = boto_vpc.associate_new_dhcp_options_to_vpc(
                vpc.id, **dhcp_options_parameters
            )
            self.assertTrue("error" in r)

    @mock_ec2_deprecated
    def test_that_when_creating_dhcp_options_set_to_a_non_existent_vpc_the_dhcp_options_the_associate_new_dhcp_options_method_returns_false(
        self,
    ):
        """
        Tests creation/association of dhcp options to non-existent vpc
        """

        r = boto_vpc.create_dhcp_options(vpc_name="fake", **dhcp_options_parameters)
        self.assertTrue("error" in r)

    @mock_ec2_deprecated
    def test_that_when_dhcp_options_exists_the_dhcp_options_exists_method_returns_true(
        self,
    ):
        """
        Tests existence of dhcp options successfully
        """
        dhcp_options = self._create_dhcp_options()

        dhcp_options_exists_result = boto_vpc.dhcp_options_exists(
            dhcp_options.id, **conn_parameters
        )

        self.assertTrue(dhcp_options_exists_result["exists"])

    @mock_ec2_deprecated
    def test_that_when_dhcp_options_do_not_exist_the_dhcp_options_exists_method_returns_false(
        self,
    ):
        """
        Tests existence of dhcp options failure
        """
        r = boto_vpc.dhcp_options_exists("fake", **conn_parameters)
        self.assertFalse(r["exists"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_checking_if_dhcp_options_exists_but_providing_no_filters_the_dhcp_options_exists_method_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests checking dhcp option existence with no filters
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following must be provided: id, name, or tags.",
        ):
            boto_vpc.dhcp_options_exists(**conn_parameters)


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
class BotoVpcNetworkACLTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_that_when_creating_network_acl_for_an_existing_vpc_the_create_network_acl_method_returns_true(
        self,
    ):
        """
        Tests creation of network acl with existing vpc
        """
        vpc = self._create_vpc()

        network_acl_creation_result = boto_vpc.create_network_acl(
            vpc.id, **conn_parameters
        )

        self.assertTrue(network_acl_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_network_acl_for_an_existing_vpc_and_specifying_a_name_the_create_network_acl_method_returns_true(
        self,
    ):
        """
        Tests creation of network acl via name with an existing vpc
        """
        vpc = self._create_vpc()

        network_acl_creation_result = boto_vpc.create_network_acl(
            vpc.id, network_acl_name="test", **conn_parameters
        )

        self.assertTrue(network_acl_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_network_acl_for_an_existing_vpc_and_specifying_tags_the_create_network_acl_method_returns_true(
        self,
    ):
        """
        Tests creation of network acl via tags with an existing vpc
        """
        vpc = self._create_vpc()

        network_acl_creation_result = boto_vpc.create_network_acl(
            vpc.id, tags={"test": "testvalue"}, **conn_parameters
        )

        self.assertTrue(network_acl_creation_result)

    @mock_ec2_deprecated
    def test_that_when_creating_network_acl_for_a_non_existent_vpc_the_create_network_acl_method_returns_an_error(
        self,
    ):
        """
        Tests creation of network acl with a non-existent vpc
        """
        network_acl_creation_result = boto_vpc.create_network_acl(
            "fake", **conn_parameters
        )

        self.assertTrue("error" in network_acl_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_network_acl_fails_the_create_network_acl_method_returns_false(
        self,
    ):
        """
        Tests creation of network acl failure
        """
        vpc = self._create_vpc()

        with patch(
            "moto.ec2.models.NetworkACLBackend.create_network_acl",
            side_effect=BotoServerError(400, "Mocked error"),
        ):
            network_acl_creation_result = boto_vpc.create_network_acl(
                vpc.id, **conn_parameters
            )

        self.assertFalse(network_acl_creation_result)

    @mock_ec2_deprecated
    def test_that_when_deleting_an_existing_network_acl_the_delete_network_acl_method_returns_true(
        self,
    ):
        """
        Tests deletion of existing network acl successfully
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)

        network_acl_deletion_result = boto_vpc.delete_network_acl(
            network_acl.id, **conn_parameters
        )

        self.assertTrue(network_acl_deletion_result)

    @mock_ec2_deprecated
    def test_that_when_deleting_a_non_existent_network_acl_the_delete_network_acl_method_returns_an_error(
        self,
    ):
        """
        Tests deleting a non-existent network acl
        """
        network_acl_deletion_result = boto_vpc.delete_network_acl(
            "fake", **conn_parameters
        )

        self.assertTrue("error" in network_acl_deletion_result)

    @mock_ec2_deprecated
    def test_that_when_a_network_acl_exists_the_network_acl_exists_method_returns_true(
        self,
    ):
        """
        Tests existence of network acl
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)

        network_acl_deletion_result = boto_vpc.network_acl_exists(
            network_acl.id, **conn_parameters
        )

        self.assertTrue(network_acl_deletion_result)

    @mock_ec2_deprecated
    def test_that_when_a_network_acl_does_not_exist_the_network_acl_exists_method_returns_false(
        self,
    ):
        """
        Tests checking network acl does not exist
        """
        network_acl_deletion_result = boto_vpc.network_acl_exists(
            "fake", **conn_parameters
        )

        self.assertFalse(network_acl_deletion_result["exists"])

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_checking_if_network_acl_exists_but_providing_no_filters_the_network_acl_exists_method_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests checking existence of network acl with no filters
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following must be provided: id, name, or tags.",
        ):
            boto_vpc.dhcp_options_exists(**conn_parameters)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_network_acl_entry_successfully_the_create_network_acl_entry_method_returns_true(
        self,
    ):
        """
        Tests creating network acl successfully
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)

        network_acl_entry_creation_result = boto_vpc.create_network_acl_entry(
            network_acl.id, *network_acl_entry_parameters, **conn_parameters
        )

        self.assertTrue(network_acl_entry_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_network_acl_entry_for_a_non_existent_network_acl_the_create_network_acl_entry_method_returns_false(
        self,
    ):
        """
        Tests creating network acl entry for non-existent network acl
        """
        network_acl_entry_creation_result = boto_vpc.create_network_acl_entry(
            *network_acl_entry_parameters, **conn_parameters
        )

        self.assertFalse(network_acl_entry_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_replacing_a_network_acl_entry_successfully_the_replace_network_acl_entry_method_returns_true(
        self,
    ):
        """
        Tests replacing network acl entry successfully
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)
        self._create_network_acl_entry(network_acl.id, *network_acl_entry_parameters)

        network_acl_entry_creation_result = boto_vpc.replace_network_acl_entry(
            network_acl.id, *network_acl_entry_parameters, **conn_parameters
        )

        self.assertTrue(network_acl_entry_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_replacing_a_network_acl_entry_for_a_non_existent_network_acl_the_replace_network_acl_entry_method_returns_false(
        self,
    ):
        """
        Tests replacing a network acl entry for a non-existent network acl
        """
        network_acl_entry_creation_result = boto_vpc.create_network_acl_entry(
            *network_acl_entry_parameters, **conn_parameters
        )
        self.assertFalse(network_acl_entry_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_an_existing_network_acl_entry_the_delete_network_acl_entry_method_returns_true(
        self,
    ):
        """
        Tests deleting existing network acl entry successfully
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)
        network_acl_entry = self._create_network_acl_entry(
            network_acl.id, *network_acl_entry_parameters
        )

        network_acl_entry_deletion_result = boto_vpc.delete_network_acl_entry(
            network_acl_entry.id, 100, **conn_parameters
        )

        self.assertTrue(network_acl_entry_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_a_non_existent_network_acl_entry_the_delete_network_acl_entry_method_returns_false(
        self,
    ):
        """
        Tests deleting a non-existent network acl entry
        """
        network_acl_entry_deletion_result = boto_vpc.delete_network_acl_entry(
            "fake", 100, **conn_parameters
        )

        self.assertFalse(network_acl_entry_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_associating_an_existing_network_acl_to_an_existing_subnet_the_associate_network_acl_method_returns_true(
        self,
    ):
        """
        Tests association of existing network acl to existing subnet successfully
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)
        subnet = self._create_subnet(vpc.id)

        network_acl_association_result = boto_vpc.associate_network_acl_to_subnet(
            network_acl.id, subnet.id, **conn_parameters
        )

        self.assertTrue(network_acl_association_result)

    @mock_ec2_deprecated
    def test_that_when_associating_a_non_existent_network_acl_to_an_existing_subnet_the_associate_network_acl_method_returns_an_error(
        self,
    ):
        """
        Tests associating a non-existent network acl to existing subnet failure
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        network_acl_association_result = boto_vpc.associate_network_acl_to_subnet(
            "fake", subnet.id, **conn_parameters
        )

        self.assertTrue("error" in network_acl_association_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_associating_an_existing_network_acl_to_a_non_existent_subnet_the_associate_network_acl_method_returns_false(
        self,
    ):
        """
        Tests associating an existing network acl to a non-existent subnet
        """
        vpc = self._create_vpc()
        network_acl = self._create_network_acl(vpc.id)

        network_acl_association_result = boto_vpc.associate_network_acl_to_subnet(
            network_acl.id, "fake", **conn_parameters
        )

        self.assertFalse(network_acl_association_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_and_associating_a_network_acl_to_a_subnet_succeeds_the_associate_new_network_acl_to_subnet_method_returns_true(
        self,
    ):
        """
        Tests creating/associating a network acl to a subnet to a new network
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        network_acl_creation_and_association_result = (
            boto_vpc.associate_new_network_acl_to_subnet(
                vpc.id, subnet.id, **conn_parameters
            )
        )

        self.assertTrue(network_acl_creation_and_association_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_and_associating_a_network_acl_to_a_subnet_and_specifying_a_name_succeeds_the_associate_new_network_acl_to_subnet_method_returns_true(
        self,
    ):
        """
        Tests creation/association of a network acl to subnet via name successfully
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        network_acl_creation_and_association_result = (
            boto_vpc.associate_new_network_acl_to_subnet(
                vpc.id, subnet.id, network_acl_name="test", **conn_parameters
            )
        )

        self.assertTrue(network_acl_creation_and_association_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_and_associating_a_network_acl_to_a_subnet_and_specifying_tags_succeeds_the_associate_new_network_acl_to_subnet_method_returns_true(
        self,
    ):
        """
        Tests creating/association of a network acl to a subnet via tag successfully
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        network_acl_creation_and_association_result = (
            boto_vpc.associate_new_network_acl_to_subnet(
                vpc.id, subnet.id, tags={"test": "testvalue"}, **conn_parameters
            )
        )

        self.assertTrue(network_acl_creation_and_association_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_and_associating_a_network_acl_to_a_non_existent_subnet_the_associate_new_network_acl_to_subnet_method_returns_false(
        self,
    ):
        """
        Tests creation/association of a network acl to a non-existent vpc
        """
        vpc = self._create_vpc()

        network_acl_creation_and_association_result = (
            boto_vpc.associate_new_network_acl_to_subnet(
                vpc.id, "fake", **conn_parameters
            )
        )

        self.assertFalse(network_acl_creation_and_association_result)

    @mock_ec2_deprecated
    def test_that_when_creating_a_network_acl_to_a_non_existent_vpc_the_associate_new_network_acl_to_subnet_method_returns_an_error(
        self,
    ):
        """
        Tests creation/association of network acl to a non-existent subnet
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        network_acl_creation_result = boto_vpc.create_network_acl(
            vpc_name="fake", subnet_id=subnet.id, **conn_parameters
        )

        self.assertTrue("error" in network_acl_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_disassociating_network_acl_succeeds_the_disassociate_network_acl_method_should_return_true(
        self,
    ):
        """
        Tests disassociation of network acl success
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        dhcp_disassociate_result = boto_vpc.disassociate_network_acl(
            subnet.id, vpc_id=vpc.id, **conn_parameters
        )

        self.assertTrue(dhcp_disassociate_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_disassociating_network_acl_for_a_non_existent_vpc_the_disassociate_network_acl_method_should_return_false(
        self,
    ):
        """
        Tests disassociation of network acl from non-existent vpc
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        dhcp_disassociate_result = boto_vpc.disassociate_network_acl(
            subnet.id, vpc_id="fake", **conn_parameters
        )

        self.assertFalse(dhcp_disassociate_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_disassociating_network_acl_for_a_non_existent_subnet_the_disassociate_network_acl_method_should_return_false(
        self,
    ):
        """
        Tests disassociation of network acl from non-existent subnet
        """
        vpc = self._create_vpc()

        dhcp_disassociate_result = boto_vpc.disassociate_network_acl(
            "fake", vpc_id=vpc.id, **conn_parameters
        )

        self.assertFalse(dhcp_disassociate_result)


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
class BotoVpcRouteTablesTestCase(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_route_table_succeeds_the_create_route_table_method_returns_true(
        self,
    ):
        """
        Tests creating route table successfully
        """
        vpc = self._create_vpc()

        route_table_creation_result = boto_vpc.create_route_table(
            vpc.id, **conn_parameters
        )

        self.assertTrue(route_table_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_route_table_on_a_non_existent_vpc_the_create_route_table_method_returns_false(
        self,
    ):
        """
        Tests creating route table on a non-existent vpc
        """
        route_table_creation_result = boto_vpc.create_route_table(
            "fake", **conn_parameters
        )

        self.assertTrue(route_table_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_a_route_table_succeeds_the_delete_route_table_method_returns_true(
        self,
    ):
        """
        Tests deleting route table successfully
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        route_table_deletion_result = boto_vpc.delete_route_table(
            route_table.id, **conn_parameters
        )

        self.assertTrue(route_table_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_a_non_existent_route_table_the_delete_route_table_method_returns_false(
        self,
    ):
        """
        Tests deleting non-existent route table
        """
        route_table_deletion_result = boto_vpc.delete_route_table(
            "fake", **conn_parameters
        )

        self.assertFalse(route_table_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_route_table_exists_the_route_table_exists_method_returns_true(
        self,
    ):
        """
        Tests existence of route table success
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        route_table_existence_result = boto_vpc.route_table_exists(
            route_table.id, **conn_parameters
        )

        self.assertTrue(route_table_existence_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_route_table_does_not_exist_the_route_table_exists_method_returns_false(
        self,
    ):
        """
        Tests existence of route table failure
        """
        route_table_existence_result = boto_vpc.route_table_exists(
            "fake", **conn_parameters
        )

        self.assertFalse(route_table_existence_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(
        reason="Disabled pending https://github.com/spulec/moto/issues/493"
    )
    def test_that_when_checking_if_a_route_table_exists_but_providing_no_filters_the_route_table_exists_method_raises_a_salt_invocation_error(
        self,
    ):
        """
        Tests checking route table without filters
        """
        with self.assertRaisesRegex(
            SaltInvocationError,
            "At least one of the following must be provided: id, name, or tags.",
        ):
            boto_vpc.dhcp_options_exists(**conn_parameters)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_associating_a_route_table_succeeds_the_associate_route_table_method_should_return_the_association_id(
        self,
    ):
        """
        Tests associating route table successfully
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)
        route_table = self._create_route_table(vpc.id)

        association_id = boto_vpc.associate_route_table(
            route_table.id, subnet.id, **conn_parameters
        )

        self.assertTrue(association_id)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_associating_a_route_table_with_a_non_existent_route_table_the_associate_route_table_method_should_return_false(
        self,
    ):
        """
        Tests associating of route table to non-existent route table
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)

        association_id = boto_vpc.associate_route_table(
            "fake", subnet.id, **conn_parameters
        )

        self.assertFalse(association_id)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_associating_a_route_table_with_a_non_existent_subnet_the_associate_route_table_method_should_return_false(
        self,
    ):
        """
        Tests associating of route table with non-existent subnet
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        association_id = boto_vpc.associate_route_table(
            route_table.id, "fake", **conn_parameters
        )

        self.assertFalse(association_id)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_disassociating_a_route_table_succeeds_the_disassociate_route_table_method_should_return_true(
        self,
    ):
        """
        Tests disassociation of a route
        """
        vpc = self._create_vpc()
        subnet = self._create_subnet(vpc.id)
        route_table = self._create_route_table(vpc.id)

        association_id = self._associate_route_table(route_table.id, subnet.id)

        dhcp_disassociate_result = boto_vpc.disassociate_route_table(
            association_id, **conn_parameters
        )

        self.assertTrue(dhcp_disassociate_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_route_succeeds_the_create_route_method_should_return_true(
        self,
    ):
        """
        Tests successful creation of a route
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        route_creation_result = boto_vpc.create_route(
            route_table.id, cidr_block, **conn_parameters
        )

        self.assertTrue(route_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_creating_a_route_with_a_non_existent_route_table_the_create_route_method_should_return_false(
        self,
    ):
        """
        Tests creation of route on non-existent route table
        """
        route_creation_result = boto_vpc.create_route(
            "fake", cidr_block, **conn_parameters
        )

        self.assertFalse(route_creation_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_a_route_succeeds_the_delete_route_method_should_return_true(
        self,
    ):
        """
        Tests deleting route from route table
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        route_deletion_result = boto_vpc.delete_route(
            route_table.id, cidr_block, **conn_parameters
        )

        self.assertTrue(route_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_deleting_a_route_with_a_non_existent_route_table_the_delete_route_method_should_return_false(
        self,
    ):
        """
        Tests deleting route from a non-existent route table
        """
        route_deletion_result = boto_vpc.delete_route(
            "fake", cidr_block, **conn_parameters
        )

        self.assertFalse(route_deletion_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_replacing_a_route_succeeds_the_replace_route_method_should_return_true(
        self,
    ):
        """
        Tests replacing route successfully
        """
        vpc = self._create_vpc()
        route_table = self._create_route_table(vpc.id)

        route_replacing_result = boto_vpc.replace_route(
            route_table.id, cidr_block, **conn_parameters
        )

        self.assertTrue(route_replacing_result)

    @mock_ec2_deprecated
    @pytest.mark.skip(reason="Moto has not implemented this feature. Skipping for now.")
    def test_that_when_replacing_a_route_with_a_non_existent_route_table_the_replace_route_method_should_return_false(
        self,
    ):
        """
        Tests replacing a route when the route table doesn't exist
        """
        route_replacing_result = boto_vpc.replace_route(
            "fake", cidr_block, **conn_parameters
        )

        self.assertFalse(route_replacing_result)


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}. Installed: {}".format(
        required_boto_version, _get_boto_version()
    ),
)
@pytest.mark.skipif(
    _has_required_moto() is False,
    reason="The moto version must be >= to version {}".format(required_moto_version),
)
class BotoVpcPeeringConnectionsTest(BotoVpcTestCaseBase, BotoVpcTestCaseMixin):
    @mock_ec2_deprecated
    def test_request_vpc_peering_connection(self):
        """
        Run with 2 vpc ids and returns a message
        """
        my_vpc = self._create_vpc()
        other_vpc = self._create_vpc()
        self.assertTrue(
            "msg"
            in boto_vpc.request_vpc_peering_connection(
                name="my_peering",
                requester_vpc_id=my_vpc.id,
                peer_vpc_id=other_vpc.id,
                **conn_parameters
            )
        )

    @mock_ec2_deprecated
    def test_request_vpc_peering_peer_region(self):
        """
        Run with 2 vpc ids with peer_region set
        and returns a message
        """
        my_vpc = self._create_vpc()
        other_vpc = self._create_vpc()
        assert "msg" in boto_vpc.request_vpc_peering_connection(
            name="my_peering",
            requester_vpc_id=my_vpc.id,
            peer_vpc_id=other_vpc.id,
            peer_region="test_region",
            **conn_parameters
        )

    @mock_ec2_deprecated
    def test_raises_error_if_both_vpc_name_and_vpc_id_are_specified(self):
        """
        Must specify only one
        """
        my_vpc = self._create_vpc()
        other_vpc = self._create_vpc()
        with self.assertRaises(SaltInvocationError):
            boto_vpc.request_vpc_peering_connection(
                name="my_peering",
                requester_vpc_id=my_vpc.id,
                requester_vpc_name="foobar",
                peer_vpc_id=other_vpc.id,
                **conn_parameters
            )

        boto_vpc.request_vpc_peering_connection(
            name="my_peering",
            requester_vpc_name="my_peering",
            peer_vpc_id=other_vpc.id,
            **conn_parameters
        )


class BotoVpcTestDeprecation(TestCase):
    def test_deprecation_58636(self):
        try:
            boto_vpc.describe_route_table
        except AttributeError:
            return
        raise AttributeError("describe_route_table should be deprecated")



import logging

import pytest

import salt.cache
import salt.loader
from tests.pytests.functional.cache.helpers import run_common_cache_tests
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


@pytest.fixture(
    scope="module",
    params=(EtcdVersion.v2, EtcdVersion.v3_v2_mode),
    ids=etcd_version_ids,
)  # pylint: disable=function-redefined
def etcd_version(request):
    # The only parameter is True because the salt cache does not use
    # salt/utils/etcd_util.py and if coded for etcd v2
    if request.param and not HAS_ETCD_V2:
        pytest.skip("No etcd library installed")
    if not request.param and not HAS_ETCD_V3:
        pytest.skip("No etcd3 library installed")
    return request.param


@pytest.fixture
def cache(minion_opts, etcd_port):
    opts = minion_opts.copy()
    opts["cache"] = "etcd"
    opts["etcd.host"] = "127.0.0.1"
    opts["etcd.port"] = etcd_port
    opts["etcd.protocol"] = "http"
    # NOTE: If you would like to ensure that alternate suffixes are properly
    # tested, simply change this value and re-run the tests.
    opts["etcd.timestamp_suffix"] = ".frobnosticate"
    cache = salt.cache.factory(opts)
    return cache


def test_caching(subtests, cache):
    run_common_cache_tests(subtests, cache)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import pytest

import salt.states.iptables as iptables
import salt.utils.state as state_utils
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        iptables: {
            "__utils__": {"state.gen_tag": state_utils.gen_tag},
        }
    }


def test_chain_present():
    """
    Test to verify the chain is exist.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(side_effect=[True, False, False, False])
    with patch.dict(iptables.__salt__, {"iptables.check_chain": mock}):
        ret.update(
            {
                "comment": (
                    "iptables salt chain is already exist in filter table for ipv4"
                )
            }
        )
        assert iptables.chain_present("salt") == ret

        with patch.dict(iptables.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": (
                        "iptables salt chain in filter"
                        " table needs to be set for ipv4"
                    ),
                    "result": None,
                }
            )
            assert iptables.chain_present("salt") == ret

        with patch.dict(iptables.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[True, ""])
            with patch.dict(iptables.__salt__, {"iptables.new_chain": mock}):
                ret.update(
                    {
                        "result": True,
                        "comment": (
                            "iptables salt chain in filter"
                            " table create success for ipv4"
                        ),
                        "changes": {"locale": "salt"},
                    }
                )
                assert iptables.chain_present("salt") == ret

                ret.update(
                    {
                        "changes": {},
                        "result": False,
                        "comment": (
                            "Failed to create salt chain in filter table:  for ipv4"
                        ),
                    }
                )
                assert iptables.chain_present("salt") == ret


def test_chain_absent():
    """
    Test to verify the chain is absent.
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(side_effect=[False, True, True, True])
    with patch.dict(iptables.__salt__, {"iptables.check_chain": mock}):
        ret.update(
            {
                "comment": (
                    "iptables salt chain is already absent in filter table for ipv4"
                )
            }
        )
        assert iptables.chain_absent("salt") == ret

        with patch.dict(iptables.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": (
                        "iptables salt chain in filter"
                        " table needs to be removed ipv4"
                    ),
                    "result": None,
                }
            )
            assert iptables.chain_absent("salt") == ret

        with patch.dict(iptables.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[False, "a"])
            with patch.dict(iptables.__salt__, {"iptables.flush": mock}):
                mock = MagicMock(return_value=True)
                with patch.dict(iptables.__salt__, {"iptables.delete_chain": mock}):
                    ret.update(
                        {
                            "changes": {"locale": "salt"},
                            "comment": (
                                "iptables salt chain in filter"
                                " table delete success for ipv4"
                            ),
                            "result": True,
                        }
                    )
                    assert iptables.chain_absent("salt") == ret

                ret.update(
                    {
                        "changes": {},
                        "result": False,
                        "comment": (
                            "Failed to flush salt chain in filter table: a for ipv4"
                        ),
                    }
                )
                assert iptables.chain_absent("salt") == ret


def test_append():
    """
    Test to append a rule to a chain
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    assert iptables.append("salt", rules=[]) == ret

    mock = MagicMock(return_value=[])
    with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value="a")
        with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
            mock = MagicMock(side_effect=[True, False, False, False, False, True])
            with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                ret.update(
                    {
                        "comment": ("iptables rule for salt already set (a) for ipv4"),
                        "result": True,
                    }
                )
                assert iptables.append("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "iptables rule for salt" " needs to be set (a) for ipv4"
                            ),
                        }
                    )
                    assert iptables.append("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": False}):
                    mock = MagicMock(side_effect=[True, False, True, True])
                    with patch.dict(iptables.__salt__, {"iptables.append": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "result": True,
                                "comment": (
                                    "Set iptables rule for salt to: a for ipv4"
                                ),
                            }
                        )
                        assert iptables.append("salt", table="", chain="") == ret

                        ret.update(
                            {
                                "changes": {},
                                "result": False,
                                "comment": (
                                    "Failed to set iptables"
                                    " rule for salt.\nAttempted rule was"
                                    " a for ipv4"
                                ),
                            }
                        )
                        assert iptables.append("salt", table="", chain="") == ret

                        mock_save = MagicMock(
                            side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                        )
                        with patch.dict(
                            iptables.__salt__, {"iptables.save": mock_save}
                        ):
                            mock_get_saved_rules = MagicMock(side_effect=[""])
                            with patch.dict(
                                iptables.__salt__,
                                {"iptables.get_saved_rules": mock_get_saved_rules},
                            ):
                                mock = MagicMock(side_effect=[""])
                                with patch.dict(
                                    iptables.__salt__, {"iptables.get_rules": mock}
                                ):
                                    ret.update(
                                        {
                                            "changes": {"locale": "salt"},
                                            "result": True,
                                            "comment": "Set and saved iptables rule"
                                            ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                        }
                                    )
                                    assert (
                                        iptables.append(
                                            "salt",
                                            table="",
                                            chain="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    ret.update(
                                        {
                                            "changes": {},
                                            "result": True,
                                            "comment": "iptables rule for salt already set (a) for ipv4",
                                        }
                                    )
                                    assert (
                                        iptables.append(
                                            "salt",
                                            table="",
                                            chain="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    assert (
                                        mock_get_saved_rules.mock_calls[0][2][
                                            "conf_file"
                                        ]
                                        == "/tmp/iptables"
                                    )
                                    assert (
                                        mock_save.mock_calls[0][2]["filename"]
                                        == "/tmp/iptables"
                                    )


def test_insert():
    """
    Test to insert a rule into a chain
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    assert iptables.insert("salt", rules=[]) == ret

    mock = MagicMock(return_value=[])
    with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value="a")
        with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
            mock = MagicMock(side_effect=[True, False, False, False, False, True])
            with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                ret.update(
                    {
                        "comment": ("iptables rule for salt already set for ipv4 (a)"),
                        "result": True,
                    }
                )
                assert iptables.insert("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "iptables rule for salt" " needs to be set for ipv4 (a)"
                            ),
                        }
                    )
                    assert iptables.insert("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": False}):
                    mock = MagicMock(side_effect=[False, True, False, True])
                    with patch.dict(iptables.__salt__, {"iptables.insert": mock}):
                        ret.update(
                            {
                                "changes": {"locale": "salt"},
                                "result": True,
                                "comment": (
                                    "Set iptables rule for salt to: a for ipv4"
                                ),
                            }
                        )
                        assert (
                            iptables.insert("salt", table="", chain="", position="")
                            == ret
                        )

                        ret.update(
                            {
                                "changes": {},
                                "result": False,
                                "comment": (
                                    "Failed to set iptables"
                                    " rule for salt.\nAttempted rule was a"
                                ),
                            }
                        )
                        assert (
                            iptables.insert("salt", table="", chain="", position="")
                            == ret
                        )

                        mock_save = MagicMock(
                            side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                        )
                        with patch.dict(
                            iptables.__salt__, {"iptables.save": mock_save}
                        ):
                            mock_get_saved_rules = MagicMock(side_effect=[""])
                            with patch.dict(
                                iptables.__salt__,
                                {"iptables.get_saved_rules": mock_get_saved_rules},
                            ):
                                mock = MagicMock(side_effect=[""])
                                with patch.dict(
                                    iptables.__salt__, {"iptables.get_rules": mock}
                                ):
                                    ret.update(
                                        {
                                            "changes": {"locale": "salt"},
                                            "result": True,
                                            "comment": "Set and saved iptables rule"
                                            ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                        }
                                    )
                                    assert (
                                        iptables.insert(
                                            "salt",
                                            table="",
                                            chain="",
                                            position="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    ret.update(
                                        {
                                            "changes": {},
                                            "result": True,
                                            "comment": "iptables rule for salt already set for ipv4 (a)",
                                        }
                                    )
                                    assert (
                                        iptables.insert(
                                            "salt",
                                            table="",
                                            chain="",
                                            position="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    assert (
                                        mock_get_saved_rules.mock_calls[0][2][
                                            "conf_file"
                                        ]
                                        == "/tmp/iptables"
                                    )
                                    assert (
                                        mock_save.mock_calls[0][2]["filename"]
                                        == "/tmp/iptables"
                                    )


def test_delete():
    """
    Test to delete a rule to a chain
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    assert iptables.delete("salt", rules=[]) == ret

    mock = MagicMock(return_value=[])
    with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value="a")
        with patch.dict(iptables.__salt__, {"iptables.build_rule": mock}):
            mock = MagicMock(side_effect=[False, True, True, True, True, False])
            with patch.dict(iptables.__salt__, {"iptables.check": mock}):
                ret.update(
                    {
                        "comment": (
                            "iptables rule for salt already absent for ipv4 (a)"
                        ),
                        "result": True,
                    }
                )
                assert iptables.delete("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": True}):
                    ret.update(
                        {
                            "result": None,
                            "comment": (
                                "iptables rule for salt needs"
                                " to be deleted for ipv4 (a)"
                            ),
                        }
                    )
                    assert iptables.delete("salt", table="", chain="") == ret

                with patch.dict(iptables.__opts__, {"test": False}):
                    mock = MagicMock(side_effect=[False, True, False, False])
                    with patch.dict(iptables.__salt__, {"iptables.delete": mock}):
                        ret.update(
                            {
                                "result": True,
                                "changes": {"locale": "salt"},
                                "comment": "Delete iptables rule for salt a",
                            }
                        )
                        assert (
                            iptables.delete("salt", table="", chain="", position="")
                            == ret
                        )

                        ret.update(
                            {
                                "result": False,
                                "changes": {},
                                "comment": (
                                    "Failed to delete iptables"
                                    " rule for salt.\nAttempted rule was a"
                                ),
                            }
                        )
                        assert (
                            iptables.delete("salt", table="", chain="", position="")
                            == ret
                        )

                        mock_save = MagicMock(
                            side_effect=['Wrote 1 lines to "/tmp/iptables"', ""]
                        )
                        with patch.dict(
                            iptables.__salt__, {"iptables.save": mock_save}
                        ):
                            mock = MagicMock(side_effect=[True, False])
                            with patch.dict(
                                iptables.__salt__, {"iptables.check": mock}
                            ):
                                mock = MagicMock(side_effect=[""])
                                with patch.dict(
                                    iptables.__salt__, {"iptables.get_rules": mock}
                                ):
                                    ret.update(
                                        {
                                            "changes": {"locale": "salt"},
                                            "result": True,
                                            "comment": "Deleted and saved iptables rule"
                                            ' salt for ipv4\na\nWrote 1 lines to "/tmp/iptables"',
                                        }
                                    )
                                    assert (
                                        iptables.delete(
                                            "salt",
                                            table="",
                                            chain="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    ret.update(
                                        {
                                            "changes": {},
                                            "result": True,
                                            "comment": "iptables rule for salt already absent for ipv4 (a)",
                                        }
                                    )
                                    assert (
                                        iptables.delete(
                                            "salt",
                                            table="",
                                            chain="",
                                            save="/tmp/iptables",
                                        )
                                        == ret
                                    )
                                    assert (
                                        mock_save.mock_calls[0][2]["filename"]
                                        == "/tmp/iptables"
                                    )


def test_set_policy():
    """
    Test to sets the default policy for iptables firewall tables
    """
    ret = {"name": "salt", "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value=[])
    with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
        mock = MagicMock(return_value="stack")
        with patch.dict(iptables.__salt__, {"iptables.get_policy": mock}):
            ret.update(
                {
                    "comment": (
                        "iptables default policy for chain"
                        "  on table  for ipv4 already set to stack"
                    )
                }
            )
            assert (
                iptables.set_policy("salt", table="", chain="", policy="stack") == ret
            )

            with patch.dict(iptables.__opts__, {"test": True}):
                ret.update(
                    {
                        "comment": (
                            "iptables default policy for chain"
                            "  on table  for ipv4 needs to be set"
                            " to sal"
                        ),
                        "result": None,
                    }
                )
                assert (
                    iptables.set_policy("salt", table="", chain="", policy="sal") == ret
                )

            with patch.dict(iptables.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[False, True])
                with patch.dict(iptables.__salt__, {"iptables.set_policy": mock}):
                    ret.update(
                        {
                            "changes": {"locale": "salt"},
                            "comment": "Set default policy for  to sal family ipv4",
                            "result": True,
                        }
                    )
                    assert (
                        iptables.set_policy("salt", table="", chain="", policy="sal")
                        == ret
                    )

                    ret.update(
                        {
                            "comment": "Failed to set iptables default policy",
                            "result": False,
                            "changes": {},
                        }
                    )
                    assert (
                        iptables.set_policy("salt", table="", chain="", policy="sal")
                        == ret
                    )


def test_flush():
    """
    Test to flush current iptables state
    """
    ret = {"name": "salt", "changes": {}, "result": None, "comment": ""}

    mock = MagicMock(return_value=[])
    with patch.object(iptables, "_STATE_INTERNAL_KEYWORDS", mock):
        with patch.dict(iptables.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": (
                        "iptables rules in salt table filter"
                        " chain ipv4 family needs to be flushed"
                    )
                }
            )
            assert iptables.flush("salt") == ret

        with patch.dict(iptables.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[False, True])
            with patch.dict(iptables.__salt__, {"iptables.flush": mock}):
                ret.update(
                    {
                        "changes": {"locale": "salt"},
                        "comment": (
                            "Flush iptables rules in  table  chain ipv4 family"
                        ),
                        "result": True,
                    }
                )
                assert iptables.flush("salt", table="", chain="") == ret

                ret.update(
                    {
                        "changes": {},
                        "comment": "Failed to flush iptables rules",
                        "result": False,
                    }
                )
                assert iptables.flush("salt", table="", chain="") == ret


def test_mod_aggregate():
    """
    Test to mod_aggregate function
    """
    low = {
        "state": "iptables",
        "name": "accept_local_interface",
        "__sls__": "iptables",
        "__env__": "base",
        "__id__": "append_accept_local_interface",
        "table": "filter",
        "chain": "INPUT",
        "in-interface": "lo",
        "jump": "ACCEPT",
        "save": True,
        "order": 10000,
        "fun": "append",
    }

    chunks = [
        {
            "state": "iptables",
            "name": "accept_local_interface",
            "__sls__": "iptables",
            "__env__": "base",
            "__id__": "append_accept_local_interface",
            "table": "filter",
            "chain": "INPUT",
            "in-interface": "lo",
            "jump": "ACCEPT",
            "save": True,
            "order": 10000,
            "fun": "append",
        },
        {
            "state": "iptables",
            "name": "append_accept_loopback_output",
            "__sls__": "iptables",
            "__env__": "base",
            "__id__": "append_accept_loopback_output",
            "table": "filter",
            "chain": "OUTPUT",
            "out-interface": "lo",
            "jump": "ACCEPT",
            "save": True,
            "order": 10001,
            "fun": "append",
        },
        {
            "state": "iptables",
            "name": "append_drop_non_loopback",
            "__sls__": "iptables",
            "__env__": "base",
            "__id__": "append_drop_non_loopback",
            "table": "filter",
            "chain": "INPUT",
            "source": "127.0.0.0/8",
            "jump": "DROP",
            "save": True,
            "order": 10002,
            "fun": "append",
        },
    ]

    expected = {
        "state": "iptables",
        "name": "accept_local_interface",
        "__sls__": "iptables",
        "__env__": "base",
        "__id__": "append_accept_local_interface",
        "table": "filter",
        "chain": "INPUT",
        "in-interface": "lo",
        "jump": "ACCEPT",
        "save": True,
        "order": 10000,
        "fun": "append",
        "rules": [
            {
                "state": "iptables",
                "name": "accept_local_interface",
                "__sls__": "iptables",
                "__env__": "base",
                "__id__": "append_accept_local_interface",
                "table": "filter",
                "chain": "INPUT",
                "in-interface": "lo",
                "jump": "ACCEPT",
                "save": True,
                "order": 10000,
                "fun": "append",
            },
            {
                "state": "iptables",
                "name": "append_accept_loopback_output",
                "__sls__": "iptables",
                "__env__": "base",
                "__id__": "append_accept_loopback_output",
                "table": "filter",
                "chain": "OUTPUT",
                "out-interface": "lo",
                "jump": "ACCEPT",
                "save": True,
                "order": 10001,
                "fun": "append",
            },
            {
                "state": "iptables",
                "name": "append_drop_non_loopback",
                "__sls__": "iptables",
                "__env__": "base",
                "__id__": "append_drop_non_loopback",
                "table": "filter",
                "chain": "INPUT",
                "source": "127.0.0.0/8",
                "jump": "DROP",
                "save": True,
                "order": 10002,
                "fun": "append",
            },
        ],
    }

    res = iptables.mod_aggregate(low, chunks, {})
    assert res == expected



import pytest

import salt.modules.win_task as task
from salt.exceptions import CommandExecutionError
from tests.support.case import ModuleCase


@pytest.mark.skip_unless_on_windows
class WinTasksTest(ModuleCase):
    """
    Tests for salt.modules.win_task.
    """

    @pytest.mark.destructive_test
    def test_adding_task_with_xml(self):
        """
        Test adding a task using xml
        """
        xml_text = r"""
        <Task version="1.2" xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task">
          <RegistrationInfo>
            <Date>2015-06-12T15:59:35.691983</Date>
            <Author>System</Author>
          </RegistrationInfo>
          <Triggers>
            <LogonTrigger>
              <Enabled>true</Enabled>
              <Delay>PT30S</Delay>
            </LogonTrigger>
          </Triggers>
          <Principals>
            <Principal id="Author">
              <UserId>System</UserId>
              <LogonType>InteractiveToken</LogonType>
              <RunLevel>HighestAvailable</RunLevel>
            </Principal>
          </Principals>
          <Settings>
            <MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy>
            <DisallowStartIfOnBatteries>true</DisallowStartIfOnBatteries>
            <StopIfGoingOnBatteries>true</StopIfGoingOnBatteries>
            <AllowHardTerminate>true</AllowHardTerminate>
            <StartWhenAvailable>false</StartWhenAvailable>
            <RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable>
            <IdleSettings>
              <StopOnIdleEnd>true</StopOnIdleEnd>
              <RestartOnIdle>false</RestartOnIdle>
            </IdleSettings>
            <AllowStartOnDemand>true</AllowStartOnDemand>
            <Enabled>true</Enabled>
            <Hidden>false</Hidden>
            <RunOnlyIfIdle>false</RunOnlyIfIdle>
            <WakeToRun>false</WakeToRun>
            <ExecutionTimeLimit>P3D</ExecutionTimeLimit>
            <Priority>4</Priority>
          </Settings>
          <Actions Context="Author">
            <Exec>
              <Command>echo</Command>
              <Arguments>"hello"</Arguments>
            </Exec>
          </Actions>
        </Task>
        """
        self.assertEqual(
            self.run_function("task.create_task_from_xml", "foo", xml_text=xml_text),
            True,
        )
        all_tasks = self.run_function("task.list_tasks")
        self.assertIn("foo", all_tasks)

    @pytest.mark.destructive_test
    def test_adding_task_with_invalid_xml(self):
        """
        Test adding a task using a malformed xml
        """
        xml_text = r"""<Malformed"""
        with self.assertRaises(CommandExecutionError):
            task.create_task_from_xml("foo", xml_text=xml_text)



"""
    Test cases for salt.modules.win_iis
"""


import pytest

import salt.modules.win_iis as win_iis
import salt.utils.json
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {win_iis: {}}


@pytest.fixture
def app_list():
    return {
        "testApp": {
            "apppool": "MyTestPool",
            "path": "/testApp",
            "preload": False,
            "protocols": ["http"],
            "sourcepath": r"C:\inetpub\apps\testApp",
        }
    }


@pytest.fixture
def apppol_list():
    return {"MyTestPool": {"applications": ["MyTestSite"], "state": "Started"}}


@pytest.fixture
def binding_list():
    return {
        "*:80:": {
            "certificatehash": None,
            "certificatestorename": None,
            "hostheader": None,
            "ipaddress": "*",
            "port": 80,
            "protocol": "http",
            "sslflags": 0,
        },
        "*:443:mytestsite.local": {
            "certificatehash": "9988776655443322111000AAABBBCCCDDDEEEFFF",
            "certificatestorename": "My",
            "hostheader": "mytestsite.local",
            "ipaddress": "*",
            "port": 443,
            "protocol": "https",
            "sslflags": 0,
        },
    }


@pytest.fixture
def site_list(binding_list):
    return {
        "MyTestSite": {
            "apppool": "MyTestPool",
            "bindings": binding_list,
            "id": 1,
            "sourcepath": r"C:\inetpub\wwwroot",
            "state": "Started",
        }
    }


@pytest.fixture
def vdir_list():
    return {"TestVdir": {"sourcepath": r"C:\inetpub\vdirs\TestVdir"}}


@pytest.fixture
def nested_vdir_list():
    return {"Test/Nested/Vdir": {"sourcepath": r"C:\inetpub\vdirs\NestedTestVdir"}}


@pytest.fixture
def list_apps_srvmgr():
    return {
        "retcode": 0,
        "stdout": salt.utils.json.dumps(
            [
                {
                    "applicationPool": "MyTestPool",
                    "name": "testApp",
                    "path": "/testApp",
                    "PhysicalPath": r"C:\inetpub\apps\testApp",
                    "preloadEnabled": False,
                    "protocols": "http",
                }
            ]
        ),
    }


@pytest.fixture
def list_apppools_srvmgr():
    return {
        "retcode": 0,
        "stdout": salt.utils.json.dumps(
            [
                {
                    "name": "MyTestPool",
                    "state": "Started",
                    "Applications": {"value": ["MyTestSite"], "Count": 1},
                }
            ]
        ),
    }


@pytest.fixture
def list_vdirs_srvmgr():
    return {
        "retcode": 0,
        "stdout": salt.utils.json.dumps(
            [{"name": "TestVdir", "physicalPath": r"C:\inetpub\vdirs\TestVdir"}]
        ),
    }


@pytest.fixture
def list_more_vdirs_srvmgr():
    return {
        "retcode": 0,
        "stdout": salt.utils.json.dumps(
            [
                {"name": "TestVdir", "physicalPath": r"C:\inetpub\vdirs\TestVdir"},
                {
                    "name": "Test/Nested/Vdir",
                    "physicalPath": r"C:\inetpub\vdirs\NestedTestVdir",
                },
            ]
        ),
    }


@pytest.fixture
def container_setting():
    return {
        "retcode": 0,
        "stdout": salt.utils.json.dumps([{"managedPipelineMode": "Integrated"}]),
    }


@pytest.fixture
def cert_binding_info():
    return "*:443:mytestsite.local"


def test_create_apppool():
    """
    Test - Create an IIS application pool.
    """
    with patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_apppools", MagicMock(return_value=dict())
    ), patch.dict(
        win_iis.__salt__
    ):
        assert win_iis.create_apppool("MyTestPool")


def test_list_apppools(apppol_list, list_apppools_srvmgr):
    """
    Test - List all configured IIS application pools.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value=list_apppools_srvmgr)
    ):
        assert win_iis.list_apppools() == apppol_list


def test_remove_apppool():
    """
    Test - Remove an IIS application pool.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_apppools",
        MagicMock(
            return_value={"MyTestPool": {"applications": list(), "state": "Started"}}
        ),
    ):
        assert win_iis.remove_apppool("MyTestPool")


def test_restart_apppool():
    """
    Test - Restart an IIS application pool.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ):
        assert win_iis.restart_apppool("MyTestPool")


def test_create_site():
    """
    Test - Create a basic website in IIS.
    """
    kwargs = {
        "name": "MyTestSite",
        "sourcepath": r"C:\inetpub\wwwroot",
        "apppool": "MyTestPool",
        "hostheader": "mytestsite.local",
        "ipaddress": "*",
        "port": 80,
        "protocol": "http",
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch("salt.modules.win_iis.list_sites", MagicMock(return_value=dict())), patch(
        "salt.modules.win_iis.list_apppools", MagicMock(return_value=dict())
    ):
        assert win_iis.create_site(**kwargs)


def test_create_site_failed():
    """
    Test - Create a basic website in IIS using invalid data.
    """
    kwargs = {
        "name": "MyTestSite",
        "sourcepath": r"C:\inetpub\wwwroot",
        "apppool": "MyTestPool",
        "hostheader": "mytestsite.local",
        "ipaddress": "*",
        "port": 80,
        "protocol": "invalid-protocol-name",
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch("salt.modules.win_iis.list_sites", MagicMock(return_value=dict())), patch(
        "salt.modules.win_iis.list_apppools", MagicMock(return_value=dict())
    ):
        pytest.raises(SaltInvocationError, win_iis.create_site, **kwargs)


def test_remove_site(site_list):
    """
    Test - Delete a website from IIS.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch("salt.modules.win_iis.list_sites", MagicMock(return_value=site_list)):
        assert win_iis.remove_site("MyTestSite")


def test_create_app(app_list):
    """
    Test - Create an IIS application.
    """
    kwargs = {
        "name": "testApp",
        "site": "MyTestSite",
        "sourcepath": r"C:\inetpub\apps\testApp",
        "apppool": "MyTestPool",
    }
    with patch.dict(win_iis.__salt__), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_apps", MagicMock(return_value=app_list)
    ):
        assert win_iis.create_app(**kwargs)


def test_list_apps(app_list, list_apps_srvmgr):
    """
    Test - Get all configured IIS applications for the specified site.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value=list_apps_srvmgr)
    ):
        assert win_iis.list_apps("MyTestSite") == app_list


def test_remove_app(app_list):
    """
    Test - Remove an IIS application.
    """
    kwargs = {"name": "otherApp", "site": "MyTestSite"}
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch("salt.modules.win_iis.list_apps", MagicMock(return_value=app_list)):
        assert win_iis.remove_app(**kwargs)


def test_create_binding(binding_list):
    """
    Test - Create an IIS binding.
    """
    kwargs = {
        "site": "MyTestSite",
        "hostheader": "",
        "ipaddress": "*",
        "port": 80,
        "protocol": "http",
        "sslflags": 0,
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_bindings", MagicMock(return_value=binding_list)
    ):
        assert win_iis.create_binding(**kwargs)


def test_create_binding_failed(binding_list):
    """
    Test - Create an IIS binding using invalid data.
    """
    kwargs = {
        "site": "MyTestSite",
        "hostheader": "",
        "ipaddress": "*",
        "port": 80,
        "protocol": "invalid-protocol-name",
        "sslflags": 999,
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_bindings", MagicMock(return_value=binding_list)
    ):
        pytest.raises(SaltInvocationError, win_iis.create_binding, **kwargs)


def test_list_bindings(binding_list, site_list):
    """
    Test - Get all configured IIS bindings for the specified site.
    """
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis.list_sites", MagicMock(return_value=site_list)
    ):
        assert win_iis.list_bindings("MyTestSite") == binding_list


def test_remove_binding(binding_list):
    """
    Test - Remove an IIS binding.
    """
    kwargs = {
        "site": "MyTestSite",
        "hostheader": "myothertestsite.local",
        "ipaddress": "*",
        "port": 443,
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_bindings", MagicMock(return_value=binding_list)
    ):
        assert win_iis.remove_binding(**kwargs)


def test_create_vdir(vdir_list):
    """
    Test - Create an IIS virtual directory.
    """
    kwargs = {
        "name": "TestVdir",
        "site": "MyTestSite",
        "sourcepath": r"C:\inetpub\vdirs\TestVdir",
    }
    with patch.dict(win_iis.__salt__), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_vdirs", MagicMock(return_value=vdir_list)
    ):
        assert win_iis.create_vdir(**kwargs)


def test_list_vdirs(list_vdirs_srvmgr):
    """
    Test - Get configured IIS virtual directories.
    """
    vdirs = {"TestVdir": {"sourcepath": r"C:\inetpub\vdirs\TestVdir"}}
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value=list_vdirs_srvmgr)
    ):
        assert win_iis.list_vdirs("MyTestSite") == vdirs


def test_remove_vdir(vdir_list):
    """
    Test - Remove an IIS virtual directory.
    """
    kwargs = {"name": "TestOtherVdir", "site": "MyTestSite"}
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch("salt.modules.win_iis.list_vdirs", MagicMock(return_value=vdir_list)):
        assert win_iis.remove_vdir(**kwargs)


def test_create_nested_vdir(nested_vdir_list):
    """
    Test - Create a nested IIS virtual directory.
    """
    kwargs = {
        "name": "Test/Nested/Vdir",
        "site": "MyTestSite",
        "sourcepath": r"C:\inetpub\vdirs\NestedTestVdir",
    }
    with patch.dict(win_iis.__salt__), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_vdirs", MagicMock(return_value=nested_vdir_list)
    ):
        assert win_iis.create_vdir(**kwargs)


def test_list_nested_vdirs(list_more_vdirs_srvmgr):
    """
    Test - Get configured IIS virtual directories.
    """
    vdirs = {
        "TestVdir": {"sourcepath": r"C:\inetpub\vdirs\TestVdir"},
        "Test/Nested/Vdir": {"sourcepath": r"C:\inetpub\vdirs\NestedTestVdir"},
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value=list_more_vdirs_srvmgr),
    ):
        assert win_iis.list_vdirs("MyTestSite") == vdirs


def test_create_cert_binding(binding_list, cert_binding_info):
    """
    Test - Assign a certificate to an IIS binding.
    """
    kwargs = {
        "name": "9988776655443322111000AAABBBCCCDDDEEEFFF",
        "site": "MyTestSite",
        "hostheader": "mytestsite.local",
        "ipaddress": "*",
        "port": 443,
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._list_certs",
        MagicMock(return_value={"9988776655443322111000AAABBBCCCDDDEEEFFF": None}),
    ), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value={"retcode": 0, "stdout": 10}),
    ), patch(
        "salt.utils.json.loads",
        MagicMock(return_value=[{"MajorVersion": 10, "MinorVersion": 0}]),
    ), patch(
        "salt.modules.win_iis.list_bindings", MagicMock(return_value=binding_list)
    ), patch(
        "salt.modules.win_iis.list_cert_bindings",
        MagicMock(return_value={cert_binding_info: binding_list[cert_binding_info]}),
    ):
        assert win_iis.create_cert_binding(**kwargs)


def test_list_cert_bindings(binding_list, site_list):
    """
    Test - List certificate bindings for an IIS site.
    """
    key = "*:443:mytestsite.local"
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis.list_sites", MagicMock(return_value=site_list)
    ):
        assert win_iis.list_cert_bindings("MyTestSite") == {key: binding_list[key]}


def test_remove_cert_binding(binding_list, cert_binding_info):
    """
    Test - Remove a certificate from an IIS binding.
    """
    kwargs = {
        "name": "FFFEEEDDDCCCBBBAAA0001112233445566778899",
        "site": "MyOtherTestSite",
        "hostheader": "myothertestsite.local",
        "ipaddress": "*",
        "port": 443,
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.list_cert_bindings",
        MagicMock(return_value={cert_binding_info: binding_list[cert_binding_info]}),
    ):
        assert win_iis.remove_cert_binding(**kwargs)


def test_get_container_setting(container_setting):
    """
    Test - Get the value of the setting for the IIS container.
    """
    kwargs = {
        "name": "MyTestSite",
        "container": "AppPools",
        "settings": ["managedPipelineMode"],
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value=container_setting)
    ):
        assert win_iis.get_container_setting(**kwargs) == {
            "managedPipelineMode": "Integrated"
        }


def test_set_container_setting():
    """
    Test - Set the value of the setting for an IIS container.
    """
    kwargs = {
        "name": "MyTestSite",
        "container": "AppPools",
        "settings": {"managedPipelineMode": "Integrated"},
    }
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._srvmgr", MagicMock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.win_iis.get_container_setting",
        MagicMock(return_value={"managedPipelineMode": "Integrated"}),
    ):
        assert win_iis.set_container_setting(**kwargs)


def test__collection_match_to_index():
    bad_match = {"key_0": "value"}
    first_match = {"key_1": "value"}
    second_match = {"key_2": "value"}
    collection = [first_match, second_match]
    settings = [{"name": "enabled", "value": collection}]
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis.get_webconfiguration_settings",
        MagicMock(return_value=settings),
    ):
        ret = win_iis._collection_match_to_index(
            "pspath", "colfilter", "name", bad_match
        )
        assert ret == -1
        ret = win_iis._collection_match_to_index(
            "pspath", "colfilter", "name", first_match
        )
        assert ret == 0
        ret = win_iis._collection_match_to_index(
            "pspath", "colfilter", "name", second_match
        )
        assert ret == 1


def test__prepare_settings():
    simple_setting = {"name": "value", "filter": "value"}
    collection_setting = {"name": "Collection[{yaml:\n\tdata}]", "filter": "value"}
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._collection_match_to_index", MagicMock(return_value=0)
    ):
        ret = win_iis._prepare_settings(
            "pspath",
            [
                simple_setting,
                collection_setting,
                {"invalid": "setting"},
                {"name": "filter-less_setting"},
            ],
        )
        assert ret == [simple_setting, collection_setting]


@patch("salt.modules.win_iis.log")
def test_get_webconfiguration_settings_empty(mock_log):
    ret = win_iis.get_webconfiguration_settings("name", settings=[])
    mock_log.warning.assert_called_once_with("No settings provided")
    assert ret == {}


def test_get_webconfiguration_settings():
    # Setup
    name = "IIS"
    collection_setting = {"name": "Collection[{yaml:\n\tdata}]", "filter": "value"}
    filter_setting = {
        "name": "enabled",
        "filter": (
            "system.webServer / security / authentication / anonymousAuthentication"
        ),
    }
    settings = [collection_setting, filter_setting]

    ps_cmd = [
        "$Settings = New-Object System.Collections.ArrayList;",
    ]
    for setting in settings:
        ps_cmd.extend(
            [
                "$Property = Get-WebConfigurationProperty -PSPath '{}'".format(name),
                "-Name '{name}' -Filter '{filter}' -ErrorAction Stop;".format(
                    filter=setting["filter"], name=setting["name"]
                ),
                "if (([String]::IsNullOrEmpty($Property) -eq $False) -and",
                "($Property.GetType()).Name -eq 'ConfigurationAttribute') {",
                "$Property = $Property | Select-Object",
                "-ExpandProperty Value };",
                "$Settings.add(@{{filter='{filter}';name='{name}';value=[String]"
                " $Property}})| Out-Null;".format(
                    filter=setting["filter"], name=setting["name"]
                ),
                "$Property = $Null;",
            ]
        )
    ps_cmd.append("$Settings")

    # Execute
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._prepare_settings", MagicMock(return_value=settings)
    ), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value={"retcode": 0, "stdout": "{}"}),
    ):
        ret = win_iis.get_webconfiguration_settings(name, settings=settings)

        # Verify
        win_iis._srvmgr.assert_called_with(cmd=ps_cmd, return_json=True)
        assert ret == {}


@patch("salt.modules.win_iis.log")
def test_set_webconfiguration_settings_empty(mock_log):
    ret = win_iis.set_webconfiguration_settings("name", settings=[])
    mock_log.warning.assert_called_once_with("No settings provided")
    assert ret is False


@patch("salt.modules.win_iis.log")
def test_set_webconfiguration_settings_no_changes(mock_log):
    # Setup
    name = "IIS"
    setting = {
        "name": "Collection[{yaml:\n\tdata}]",
        "filter": (
            "system.webServer / security / authentication / anonymousAuthentication"
        ),
        "value": [],
    }
    settings = [setting]

    # Execute
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._prepare_settings", MagicMock(return_value=settings)
    ), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value={"retcode": 0, "stdout": "{}"}),
    ), patch(
        "salt.modules.win_iis.get_webconfiguration_settings",
        MagicMock(return_value=settings),
    ):
        ret = win_iis.set_webconfiguration_settings(name, settings=settings)

        # Verify
        mock_log.debug.assert_called_with(
            "Settings already contain the provided values."
        )
        assert ret is True


@patch("salt.modules.win_iis.log")
def test_set_webconfiguration_settings_failed(mock_log):
    # Setup
    name = "IIS"
    setting = {
        "name": "Collection[{yaml:\n\tdata}]",
        "filter": (
            "system.webServer / security / authentication / anonymousAuthentication"
        ),
        "value": [],
    }
    settings = [setting]

    # Execute
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._prepare_settings", MagicMock(return_value=settings)
    ), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value={"retcode": 0, "stdout": "{}"}),
    ), patch(
        "salt.modules.win_iis.get_webconfiguration_settings",
        MagicMock(side_effect=[[], [{"value": "unexpected_change!"}]]),
    ):

        ret = win_iis.set_webconfiguration_settings(name, settings=settings)

        # Verify
        assert ret is False
        mock_log.error.assert_called_with("Failed to change settings: %s", settings)


@patch("salt.modules.win_iis.log")
def test_set_webconfiguration_settings(mock_log):
    # Setup
    name = "IIS"
    setting = {
        "name": "Collection[{yaml:\n\tdata}]",
        "filter": (
            "system.webServer / security / authentication / anonymousAuthentication"
        ),
        "value": [],
    }
    settings = [setting]

    # Execute
    with patch.dict(win_iis.__salt__), patch(
        "salt.modules.win_iis._prepare_settings", MagicMock(return_value=settings)
    ), patch(
        "salt.modules.win_iis._srvmgr",
        MagicMock(return_value={"retcode": 0, "stdout": "{}"}),
    ), patch(
        "salt.modules.win_iis.get_webconfiguration_settings",
        MagicMock(side_effect=[[], settings]),
    ):
        ret = win_iis.set_webconfiguration_settings(name, settings=settings)

        # Verify
        assert ret is True
        mock_log.debug.assert_called_with(
            "Settings configured successfully: %s", settings
        )


def test_get_webconfiguration_settings_no_settings():
    assert win_iis.get_webconfiguration_settings("salt", {}) == {}


def test_get_webconfiguration_settings_pass():
    settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
        }
    ]

    ps_cmd_validate = [
        "Get-WebConfigurationProperty",
        "-PSPath",
        "'salt'",
        "-Filter",
        "'system.webServer/security/authentication/anonymousAuthentication'",
        "-Name",
        "'enabled'",
        "-ErrorAction",
        "Stop",
        "|",
        "Out-Null;",
    ]

    ps_cmd = [
        "$Settings = New-Object System.Collections.ArrayList;",
        "$Property = Get-WebConfigurationProperty -PSPath 'salt'",
        "-Name 'enabled' -Filter"
        " 'system.webServer/security/authentication/anonymousAuthentication'"
        " -ErrorAction Stop;",
        "if (([String]::IsNullOrEmpty($Property) -eq $False) -and",
        "($Property.GetType()).Name -eq 'ConfigurationAttribute') {",
        "$Property = $Property | Select-Object",
        "-ExpandProperty Value };",
        "$Settings.add(@{filter='system.webServer/security/authentication/anonymousAuthentication';name='enabled';value=[String]"
        " $Property})| Out-Null;",
        "$Property = $Null;",
        "$Settings",
    ]

    func_ret = {"name": "enabled", "value": True}
    with patch.object(
        win_iis, "_srvmgr", return_value={"retcode": 0, "stdout": "json data"}
    ) as _srvmgr:
        with patch.object(
            win_iis.salt.utils.json, "loads", return_value=func_ret
        ) as loads:
            ret = win_iis.get_webconfiguration_settings("salt", settings)

            assert _srvmgr.call_count == 2
            assert _srvmgr.mock_calls[0] == call(cmd=ps_cmd_validate, return_json=True)
            assert _srvmgr.mock_calls[1] == call(cmd=ps_cmd, return_json=True)

            loads.assert_called_once_with("json data", strict=False)
            assert func_ret == ret


def test_set_webconfiguration_settings_no_settings():
    assert win_iis.set_webconfiguration_settings("salt", {}) is False


def test_set_webconfiguration_settings_pass():
    settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": False,
        }
    ]

    current_settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": True,
        }
    ]

    new_settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": False,
        }
    ]

    ps_cmd = [
        "Set-WebConfigurationProperty",
        "-PSPath",
        "'salt'",
        "-Filter",
        "'system.webServer/security/authentication/anonymousAuthentication'",
        "-Name",
        "'enabled'",
        "-Value",
        "'False';",
    ]

    with patch.object(
        win_iis,
        "get_webconfiguration_settings",
        side_effect=[current_settings, new_settings],
    ) as get_webconfiguration_settings:
        with patch.object(win_iis, "_srvmgr", return_value={"retcode": 0}) as _srvmgr:
            ret = win_iis.set_webconfiguration_settings("salt", settings)

            assert get_webconfiguration_settings.call_count == 2
            assert get_webconfiguration_settings.mock_calls[0] == call(
                name="salt", settings=settings
            )
            assert get_webconfiguration_settings.mock_calls[1] == call(
                name="salt", settings=settings
            )

            _srvmgr.assert_called_once_with(ps_cmd)

            assert ret


def test_set_webconfiguration_settings_fail():
    settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": False,
        }
    ]

    current_settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": True,
        }
    ]

    new_settings = [
        {
            "name": "enabled",
            "filter": (
                "system.webServer/security/authentication/anonymousAuthentication"
            ),
            "value": True,
        }
    ]

    ps_cmd = [
        "Set-WebConfigurationProperty",
        "-PSPath",
        "'salt'",
        "-Filter",
        "'system.webServer/security/authentication/anonymousAuthentication'",
        "-Name",
        "'enabled'",
        "-Value",
        "'False';",
    ]

    with patch.object(
        win_iis,
        "get_webconfiguration_settings",
        side_effect=[current_settings, new_settings],
    ) as get_webconfiguration_settings:
        with patch.object(win_iis, "_srvmgr", return_value={"retcode": 0}) as _srvmgr:
            ret = win_iis.set_webconfiguration_settings("salt", settings)

            assert get_webconfiguration_settings.call_count == 2
            assert get_webconfiguration_settings.mock_calls[0] == call(
                name="salt", settings=settings
            )
            assert get_webconfiguration_settings.mock_calls[1] == call(
                name="salt", settings=settings
            )

            _srvmgr.assert_called_once_with(ps_cmd)

            assert not ret



"""
Package support for OpenBSD

.. note::

    The package repository is configured on each host using ``/etc/installurl``
    from OpenBSD 6.1 onwards. Earlier releases relied on ``/etc/pkg.conf``.

.. versionchanged:: 2016.3.5

    Package versions on OpenBSD are not normally specified explicitly; instead
    packages may be available in multiple *flavors*, and *branches* which are
    specified by the format of the package name. This module allows you to use
    the same formatting as ``pkg_add(1)``, and will select the empty flavor and
    default branch by default. Examples:

    .. code-block:: yaml

      - rsync
      - vim--no_x11
      - ruby%2.3

"""

import copy
import logging
import re

import salt.utils.data
import salt.utils.versions
from salt.exceptions import CommandExecutionError, MinionError

log = logging.getLogger(__name__)

# FIXME: replace guesswork with `pkg_info -z` to correctly identify package
#        flavors and branches
__PKG_RE = re.compile("^((?:[^-]+|-(?![0-9]))+)-([0-9][^-]*)(?:-(.*))?$")

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Set the virtual pkg module if the os is OpenBSD
    """
    if __grains__["os"] == "OpenBSD":
        return __virtualname__
    return (
        False,
        "The openbsdpkg execution module cannot be loaded: "
        "only available on OpenBSD systems.",
    )


def _list_pkgs_from_context(versions_as_list):
    """
    Use pkg list from __context__
    """
    if versions_as_list:
        return __context__["pkg.list_pkgs"]
    else:
        ret = copy.deepcopy(__context__["pkg.list_pkgs"])
        __salt__["pkg_resource.stringify"](ret)
        return ret


def list_pkgs(versions_as_list=False, **kwargs):
    """
    List the packages currently installed as a dict::

        {'<package_name>': '<version>'}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.list_pkgs
    """
    versions_as_list = salt.utils.data.is_true(versions_as_list)
    # not yet implemented or not applicable
    if any(
        [salt.utils.data.is_true(kwargs.get(x)) for x in ("removed", "purge_desired")]
    ):
        return {}

    if "pkg.list_pkgs" in __context__ and kwargs.get("use_context", True):
        return _list_pkgs_from_context(versions_as_list)

    ret = {}
    cmd = "pkg_info -q -a"
    out = __salt__["cmd.run_stdout"](cmd, output_loglevel="trace")
    for line in out.splitlines():
        try:
            pkgname, pkgver, flavor = __PKG_RE.match(line).groups()
        except AttributeError:
            continue
        pkgname += "--{}".format(flavor) if flavor else ""
        __salt__["pkg_resource.add_pkg"](ret, pkgname, pkgver)

    __salt__["pkg_resource.sort_pkglist"](ret)
    __context__["pkg.list_pkgs"] = copy.deepcopy(ret)
    if not versions_as_list:
        __salt__["pkg_resource.stringify"](ret)
    return ret


def latest_version(*names, **kwargs):
    """
    Return the latest version of the named package available for upgrade or
    installation. If more than one package name is specified, a dict of
    name/version pairs is returned.

    If the latest version of a given package is already installed, an empty
    string will be returned for that package.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.latest_version <package name>
    """
    kwargs.pop("refresh", True)

    pkgs = list_pkgs()
    ret = {}
    # Initialize the dict with empty strings
    for name in names:
        ret[name] = ""

        # Query the repository for the package name
        cmd = "pkg_info -Q {}".format(name)
        out = __salt__["cmd.run_stdout"](
            cmd, python_shell=False, output_loglevel="trace"
        )

        # Since we can only query instead of request the specific package
        # we'll have to go through the returned list and find what we
        # were looking for.
        # Keep in mind the match may be flavored.
        for line in out.splitlines():
            try:
                pkgname, pkgver, flavor = __PKG_RE.match(line).groups()
            except AttributeError:
                continue

            match = re.match(r".*\(installed\)$", pkgver)
            if match:
                # Package is explicitly marked as installed already,
                # so skip any further comparison and move on to the
                # next package to compare (if provided).
                break

            # First check if we need to look for flavors before
            # looking at unflavored packages.
            if "{}--{}".format(pkgname, flavor) == name:
                pkgname += "--{}".format(flavor)
            elif pkgname == name:
                pass
            else:
                # No match just move on.
                continue

            cur = pkgs.get(pkgname, "")
            if not cur or salt.utils.versions.compare(ver1=cur, oper="<", ver2=pkgver):
                ret[pkgname] = pkgver

    # Return a string if only one package name passed
    if len(names) == 1:
        return ret[names[0]]
    return ret


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    """
    return __salt__["pkg_resource.version"](*names, **kwargs)


def install(name=None, pkgs=None, sources=None, **kwargs):
    """
    Install the passed package

    Return a dict containing the new package names and versions::

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example, Install one package:

    .. code-block:: bash

        salt '*' pkg.install <package name>

    CLI Example, Install more than one package:

    .. code-block:: bash

        salt '*' pkg.install pkgs='["<package name>", "<package name>"]'

    CLI Example, Install more than one package from a alternate source (e.g.
    salt file-server, HTTP, FTP, local filesystem):

    .. code-block:: bash

        salt '*' pkg.install sources='[{"<pkg name>": "salt://pkgs/<pkg filename>"}]'
    """
    try:
        pkg_params, pkg_type = __salt__["pkg_resource.parse_targets"](
            name, pkgs, sources, **kwargs
        )
    except MinionError as exc:
        raise CommandExecutionError(exc)

    if not pkg_params:
        return {}

    old = list_pkgs()
    errors = []
    for pkg in pkg_params:
        # A special case for OpenBSD package "branches" is also required in
        # salt/states/pkg.py
        if pkg_type == "repository":
            stem, branch = (pkg.split("%") + [""])[:2]
            base, flavor = (stem.split("--") + [""])[:2]
            pkg = "{}--{}%{}".format(base, flavor, branch)
        cmd = "pkg_add -x -I {}".format(pkg)
        out = __salt__["cmd.run_all"](cmd, python_shell=False, output_loglevel="trace")
        if out["retcode"] != 0 and out["stderr"]:
            errors.append(out["stderr"])

    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problem encountered installing package(s)",
            info={"errors": errors, "changes": ret},
        )

    return ret


def remove(name=None, pkgs=None, purge=False, **kwargs):
    """
    Remove a single package with pkg_delete

    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.remove <package name>
        salt '*' pkg.remove <package1>,<package2>,<package3>
        salt '*' pkg.remove pkgs='["foo", "bar"]'
    """
    try:
        pkg_params = [
            x.split("--")[0]
            for x in __salt__["pkg_resource.parse_targets"](name, pkgs)[0]
        ]
    except MinionError as exc:
        raise CommandExecutionError(exc)

    old = list_pkgs()
    targets = [x for x in pkg_params if x in old]
    if not targets:
        return {}

    cmd = ["pkg_delete", "-Ix", "-Ddependencies"]

    if purge:
        cmd.append("-cqq")

    cmd.extend(targets)

    out = __salt__["cmd.run_all"](cmd, python_shell=False, output_loglevel="trace")
    if out["retcode"] != 0 and out["stderr"]:
        errors = [out["stderr"]]
    else:
        errors = []

    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if errors:
        raise CommandExecutionError(
            "Problem encountered removing package(s)",
            info={"errors": errors, "changes": ret},
        )

    return ret


def purge(name=None, pkgs=None, **kwargs):
    """
    Remove a package and extra configuration files.

    name
        The name of the package to be deleted.


    Multiple Package Options:

    pkgs
        A list of packages to delete. Must be passed as a python list. The
        ``name`` parameter will be ignored if this option is passed.

    .. versionadded:: 0.16.0


    Returns a dict containing the changes.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.purge <package name>
        salt '*' pkg.purge <package1>,<package2>,<package3>
        salt '*' pkg.purge pkgs='["foo", "bar"]'
    """
    return remove(name=name, pkgs=pkgs, purge=True)


def upgrade_available(name, **kwargs):
    """
    Check whether or not an upgrade is available for a given package

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade_available <package name>
    """
    return latest_version(name) != ""


def upgrade(name=None, pkgs=None, **kwargs):
    """
    Run a full package upgrade (``pkg_add -u``), or upgrade a specific package
    if ``name`` or ``pkgs`` is provided.
    ``name`` is ignored when ``pkgs`` is specified.

    Returns a dictionary containing the changes:

    .. versionadded:: 2019.2.0

    .. code-block:: python

        {'<package>': {'old': '<old-version>',
                       'new': '<new-version>'}}

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.upgrade
        salt '*' pkg.upgrade python%2.7
    """
    old = list_pkgs()

    cmd = ["pkg_add", "-Ix", "-u"]

    if kwargs.get("noop", False):
        cmd.append("-n")

    if pkgs:
        cmd.extend(pkgs)
    elif name:
        cmd.append(name)

    # Now run the upgrade, compare the list of installed packages before and
    # after and we have all the info we need.
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    __context__.pop("pkg.list_pkgs", None)
    new = list_pkgs()
    ret = salt.utils.data.compare_dicts(old, new)

    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Problem encountered upgrading packages",
            info={"changes": ret, "result": result},
        )

    return ret



"""
    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>
"""
import pytest

import salt.modules.libcloud_dns as libcloud_dns
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock
from tests.support.unit import TestCase


class MockDNSDriver:
    def __init__(self):
        pass


def get_mock_driver():
    return MockDNSDriver()


@pytest.mark.skipif(not libcloud_dns.HAS_LIBCLOUD, reason="No libcloud installed")
class LibcloudDnsModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "_get_driver": MagicMock(return_value=MockDNSDriver()),
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                )
            },
        }
        if libcloud_dns.HAS_LIBCLOUD is False:
            module_globals["sys.modules"] = {"libcloud": MagicMock()}

        return {libcloud_dns: module_globals}

    def test_module_creation(self):
        client = libcloud_dns._get_driver("test")
        self.assertFalse(client is None)



"""
Manage modjk workers
====================

Send commands to a :strong:`modjk` load balancer via the peer system.

This module can be used with the :ref:`prereq <requisites-prereq>`
requisite to remove/add the worker from the load balancer before
deploying/restarting service.

Mandatory Settings:

- The minion needs to have permission to publish the :strong:`modjk.*`
  functions (see :ref:`here <peer>` for information on configuring
  peer publishing permissions)

- The modjk load balancer must be configured as stated in the :strong:`modjk`
  execution module :mod:`documentation <salt.modules.modjk>`
"""


def __virtual__():
    """
    Check if we have peer access ?
    """
    return True


def _send_command(cmd, worker, lbn, target, profile="default", tgt_type="glob"):
    """
    Send a command to the modjk loadbalancer
    The minion need to be able to publish the commands to the load balancer

    cmd:
        worker_stop - won't get any traffic from the lbn
        worker_activate - activate the worker
        worker_disable - will get traffic only for current sessions
    """

    ret = {
        "code": False,
        "msg": "OK",
        "minions": [],
    }

    # Send the command to target
    func = "modjk.{}".format(cmd)
    args = [worker, lbn, profile]
    response = __salt__["publish.publish"](target, func, args, tgt_type)

    # Get errors and list of affeced minions
    errors = []
    minions = []
    for minion in response:
        minions.append(minion)
        if not response[minion]:
            errors.append(minion)

    # parse response
    if not response:
        ret["msg"] = "no servers answered the published command {}".format(cmd)
        return ret
    elif len(errors) > 0:
        ret["msg"] = "the following minions return False"
        ret["minions"] = errors
        return ret
    else:
        ret["code"] = True
        ret["msg"] = "the commad was published successfully"
        ret["minions"] = minions
        return ret


def _worker_status(target, worker, activation, profile="default", tgt_type="glob"):
    """
    Check if the worker is in `activation` state in the targeted load balancers

    The function will return the following dictionary:
        result - False if no server returned from the published command
        errors - list of servers that couldn't find the worker
        wrong_state - list of servers that the worker was in the wrong state
                      (not activation)
    """

    ret = {
        "result": True,
        "errors": [],
        "wrong_state": [],
    }

    args = [worker, profile]
    status = __salt__["publish.publish"](target, "modjk.worker_status", args, tgt_type)

    # Did we got any respone from someone ?
    if not status:
        ret["result"] = False
        return ret

    # Search for errors & status
    for balancer in status:
        if not status[balancer]:
            ret["errors"].append(balancer)
        elif status[balancer]["activation"] != activation:
            ret["wrong_state"].append(balancer)

    return ret


def _talk2modjk(name, lbn, target, action, profile="default", tgt_type="glob"):
    """
    Wrapper function for the stop/disable/activate functions
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    action_map = {
        "worker_stop": "STP",
        "worker_disable": "DIS",
        "worker_activate": "ACT",
    }

    # Check what needs to be done
    status = _worker_status(target, name, action_map[action], profile, tgt_type)
    if not status["result"]:
        ret["result"] = False
        ret["comment"] = "no servers answered the published command modjk.worker_status"
        return ret
    if status["errors"]:
        ret["result"] = False
        ret[
            "comment"
        ] = "the following balancers could not find the worker {}: {}".format(
            name, status["errors"]
        )
        return ret
    if not status["wrong_state"]:
        ret[
            "comment"
        ] = "the worker is in the desired activation state on all the balancers"
        return ret
    else:
        ret["comment"] = "the action {} will be sent to the balancers {}".format(
            action, status["wrong_state"]
        )
        ret["changes"] = {action: status["wrong_state"]}

    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Send the action command to target
    response = _send_command(action, name, lbn, target, profile, tgt_type)
    ret["comment"] = response["msg"]
    ret["result"] = response["code"]
    return ret


def stop(name, lbn, target, profile="default", tgt_type="glob"):
    """
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Stop the named worker from the lbn load balancers at the targeted minions
    The worker won't get any traffic from the lbn

    Example:

    .. code-block:: yaml

        disable-before-deploy:
          modjk_worker.stop:
            - name: {{ grains['id'] }}
            - lbn: application
            - target: 'roles:balancer'
            - tgt_type: grain
    """
    return _talk2modjk(name, lbn, target, "worker_stop", profile, tgt_type)


def activate(name, lbn, target, profile="default", tgt_type="glob"):
    """
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Activate the named worker from the lbn load balancers at the targeted
    minions

    Example:

    .. code-block:: yaml

        disable-before-deploy:
          modjk_worker.activate:
            - name: {{ grains['id'] }}
            - lbn: application
            - target: 'roles:balancer'
            - tgt_type: grain
    """
    return _talk2modjk(name, lbn, target, "worker_activate", profile, tgt_type)


def disable(name, lbn, target, profile="default", tgt_type="glob"):
    """
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Disable the named worker from the lbn load balancers at the targeted
    minions. The worker will get traffic only for current sessions and won't
    get new ones.

    Example:

    .. code-block:: yaml

        disable-before-deploy:
          modjk_worker.disable:
            - name: {{ grains['id'] }}
            - lbn: application
            - target: 'roles:balancer'
            - tgt_type: grain
    """
    return _talk2modjk(name, lbn, target, "worker_disable", profile, tgt_type)
